import os
import pathlib
from typing import Callable, Dict, List, Optional, Union

from utipy.path.mk_rm_dir import mk_dir, rm_dir as remove_dir
from utipy.path.prepare_paths import prepare_in_out_paths
from utipy.utils.messenger import Messenger


# TODO Add tests
# TODO Allow keys to have a list of paths?
# TODO Allow checking if dirs are empty (e.g., checkpoints)


class IOPaths:
    COLLECTION_NAMES = [
        "in_files",
        "in_dirs",
        "out_files",
        "out_dirs",
        "tmp_files",
        "tmp_dirs",
    ]

    def __init__(
        self,
        in_files: Dict[str, Union[str, pathlib.PurePath]] = None,
        in_dirs: Dict[str, Union[str, pathlib.PurePath]] = None,
        out_files: Dict[str, Union[str, pathlib.PurePath]] = None,
        out_dirs: Dict[str, Union[str, pathlib.PurePath]] = None,
        tmp_files: Dict[str, Union[str, pathlib.PurePath]] = None,
        tmp_dirs: Dict[str, Union[str, pathlib.PurePath]] = None,
        allow_none: bool = False,
        allow_overwriting: bool = True,
        allow_duplicates_in: List[str] = ["in_dirs", "out_dirs", "tmp_dirs"],
        disallowed_nestings: List[tuple] = [
            ("in_files", "tmp_dirs"),
            ("out_files", "tmp_dirs"),
            ("in_dirs", "tmp_dirs"),
            ("out_dirs", "tmp_dirs"),
        ],
        print_note: bool = "",
    ) -> None:
        """
        Collection of path collections for keeping track of in- and output paths,
        be it files or directories. Has relevant checks, such as duplication and
        existence, and allows printing a summary of the paths. Paths are converted to
        `pathlib.Path` objects and missing directories can be created.

        Note: Even when `allow_overwriting` is enabled,
        `in_files` and `out_files` cannot contain the same paths.

        Keys must be unique across the path dictionaries.

        Parameters
        ----------
        in_files : dict
            Dict with named paths to input files, mapping `(argument) name -> input filepath`.
            A path can also be "-" when streaming input.
            This value will remain a string and will not be checked
            for duplication, existence, etc.
        in_dirs : dict
            Dict with named paths to input directories, mapping `(argument) name -> input directory path`.
        out_files : dict
            Dict with named paths to output files, mapping `(argument) name -> output filepath`.
        out_dirs : dict
            Dict with named paths to output directories, mapping `(argument) name -> output directory path`.
        tmp_files : dict
            Dict with named paths to temporary files, mapping `(argument) name -> temporary filepath`.
        tmp_dirs : dict
            Dict with named paths to temporary directories, mapping `(argument) name -> temporary directory path`.
        allow_none : bool
            Whether to allow paths (i.e. dict values) to be `None`,
            in which case they will be ignored.
        allow_overwriting : bool
            Whether to allow `out_files` paths to already exist.
        allow_duplicates_in : list
            List of collections to allow duplicate paths in. One of:
                {'in_files', 'in_dirs', 'out_files', 'out_dirs', 'tmp_files', 'tmp_dirs'}.
        disallowed_nestings : list of tuples
            Pairs of collections where the paths of the first cannot
            be nested inside the paths of the second.
            By default, the in and out paths cannot be located within
            the temporary directories, as these might be deleted.
        print_note : string
            String to add to end of `__str__` method.
            That is a suffix added when printing the collection.

        Examples
        --------

        Initialize the collection of path collections.
        Usually the paths come from input arguments or variables.
        Additional paths can be added later with `set_path()`.

        >>> paths = IOPaths(
        ...     in_files={
        ...         "in_file": "../dir1/dir2/john.csv",
        ...         "stream_in": "-"
        ...     },
        ...     in_dirs={
        ...         "in_dir": "../dir1/dir2/",
        ...     },
        ...     out_files={
        ...         "out_file": "../dir1/dir2/output/no_john.csv"
        ...     },
        ...     out_dirs={
        ...         "out_path": "../dir1/dir2/output/"
        ...     }
        ... )

        Add additional path. Reruns checks to ensure consistency.

        >>> paths.set_path(
        ...     name="in_file_2",
        ...     path="../dir1/dir2/dennis.csv",
        ...     collection="in_files"
        ... )

        Or set multiple at a time to avoid rerunning checks unnecessarily.

        >>> paths.set_paths(
        ...     paths={
        ...         "out_file_2": "../dir1/dir2/output/no_dennis.csv"
        ...         "out_file_3": "../dir1/dir2/output/readme.txt"
        ...     },
        ...     collection="out_files"
        ... )

        Create the output directories that do not exist.

        >>> paths.mk_output_dirs(collection="out_dirs")

        Get a path.

        >>> paths["in_file"]  # or
        >>> paths.get_path(name="in_file", as_str=False, raise_on_fail=True)

        Remove file from disk.

        >>> paths.rm_file(name="in_file")

        Update collection with another `IOPaths` collection.
        The sub collections are dicts, why this is just dict.update()
        on each sub collection.

        >>> paths.update(other=other_paths)

        Find the combinations of keys and paths in the collections of this object
        that are not in the collections of another object.

        >>> paths.difference(other=other_paths):
        """
        self.all_paths = None
        self._collections = {
            "in_files": in_files,
            "in_dirs": in_dirs,
            "out_files": out_files,
            "out_dirs": out_dirs,
            "tmp_files": tmp_files,
            "tmp_dirs": tmp_dirs,
        }

        self.allow_none = allow_none
        self.allow_overwriting = allow_overwriting
        self.allow_duplicates_in = allow_duplicates_in
        self.disallowed_nestings = disallowed_nestings
        self.print_note = print_note

        self._prepare_paths()

    # Properties

    @property
    def npaths(self):
        """
        Total number of stored paths.

        Returns
        -------
        Int
        """
        return sum(self.collection_sizes().values())

    @property
    def collection_sizes(self):
        """
        Sizes of collections (size: number of paths).

        Returns
        -------
        Dict mapping collection name -> number of paths.
        """
        return {
            coll: self.get_collection_size(coll) for coll in IOPaths.COLLECTION_NAMES
        }

    def get_collection_size(self, name: str):
        """
        Get size (number of paths) of a given collection.

        Parameters
        ----------
        name : str
            Name of collection to get size of.

        Returns
        -------
        Int
        """
        coll = self.get_collection(name)
        if coll is None:
            return 0
        return len(coll)

    # Getters and setters

    def get_collection(self, name: str):
        """
        Get collection of paths.

        Parameters
        ----------
        name : str
            Name of collection to get.

        Returns
        -------
        Dict mapping `name -> path`.
        """
        assert isinstance(name, str)
        self._check_collection_name(name)
        return self._collections[name]

    def set_collection(self, name: str, collection: dict):
        """
        Set collection of paths.

        See also: `update()`.

        Parameters
        ----------
        name : str
            Name of collection to set.
        collection : dict
            Collection of paths as dict, mapping `name -> path`.
            Replaces existing dict.
        """
        # NOTE: Avoid internal use
        # as it may cause an infinite loop with
        # `self._prepare_paths()`
        self._set_collection(name=name, coll=collection)
        # Ensure consistency
        self._prepare_paths()

    def _set_collection(self, name: str, coll: Union[dict, None]):
        assert isinstance(name, str)
        assert coll is None or isinstance(coll, dict)
        if name not in IOPaths.COLLECTION_NAMES:
            raise ValueError(
                f"`name` was not one of the allowed collection names: {IOPaths.COLLECTION_NAMES}"
            )
        self._collections[name] = coll

    def _update_collection(self, paths: dict, collection: str):
        assert isinstance(paths, dict)
        assert isinstance(collection, str)
        self._check_collection_name(collection)
        if self._collections[collection] is None:
            self._set_collection(name=collection, coll=paths)
        self._collections[collection].update(paths)

    def __getitem__(self, name: str):
        """
        Get path from its key:

        e.g., d['key']
        """
        return self.get_path(name=name)

    def get_path(
        self, name: str, as_str: bool = False, raise_on_fail: bool = True
    ) -> Optional[Union[pathlib.Path, str]]:
        """
        Get a path.

        Parameters
        ----------
        name : str
            Name of path to get.
        as_str : bool
            Whether to convert the path to string (default is as `pathlib.Path`).
        raise_on_fail : bool
            Whether to raise an error when the path does not exist.
            When `False`, it defaults to `None`.

        Returns
        -------
        `pathlib.Path`, str or `None`
        """
        if name in self.all_paths:
            return self._format_path_out(path=self.all_paths[name], as_str=as_str)
        if raise_on_fail:
            raise ValueError(
                f"{name} was not a known key in any of the path collections."
            )
        return None

    def set_path(
        self, name: str, path: Union[str, pathlib.PurePath], collection: str
    ) -> None:
        """
        Set a path in a collection.

        Parameters
        ----------
        name : str
            Name of path for collection dict.
        path : str or `pathlib.Path`
            The path to set.
        collection : str
            Name of collection to set the path in.
            When a name already exist it is overwritten.
        """
        # NOTE: Avoid internal use
        # as it may cause an infinite loop
        assert isinstance(name, str)
        assert isinstance(path, (str, pathlib.PurePath))
        self._update_collection(paths={name: path}, collection=collection)

        # Ensure consistency
        self._prepare_paths()

    def set_paths(self, paths: dict, collection: str) -> None:
        """
        Set one or more paths in a collection.

        Parameters
        ----------
        paths : dict
            Dict mapping `path name -> path`.
            Paths can be of type str or `pathlib.Path`.
        collection : str
            Name of collection to set the path in.
        """
        # NOTE: Avoid internal use
        # as it may cause an infinite loop
        self._update_collection(paths=paths, collection=collection)

        # Ensure consistency
        self._prepare_paths()

    def rm_path(self, name: str) -> None:
        """
        Remove path from collection.

        Parameters
        ----------
        name : str
            Name of path in collection.
        """
        coll_name = self.get_collection_name_by_path_name(name)
        del self._collections[coll_name][name]
        del self.all_paths[name]

    def rm_paths(self, names: List[str]) -> None:
        """
        Remove paths from collections.

        Parameters
        ----------
        names : list of str
            Names of paths in one or more collections.
        """
        [self.rm_path(name=name) for name in names]

    def rm_paths_in_dir(
        self,
        dir_path_name: Optional[str] = None,
        dir_path: Optional[Union[str, pathlib.PurePath]] = None,
        rm_dir_path: bool = True,
    ) -> None:
        """
        Remove paths from collection that are within a directory.

        Directory can be specified either as the name of a path in the collections
        or a new path. The directory does not need to exist.

        Parameters
        ----------
        dir_path_name : str or `None`
            Name of path in the collections.
        dir_path : str or `pathlib.Path` or `None`
            A path to a directory.
            No checks are done on this path but it is
            resolved with `pathlib.Path.resolve()`.
        rm_dir_path : bool
            Whether to remove the path to the directory itself.
        """
        if sum([dir_path_name is not None, dir_path is not None]) != 1:
            raise ValueError(
                "Exactly one of {`dir_path_name`, `dir_path`} should be specified."
            )
        if dir_path_name is not None:
            dir_path = self[dir_path_name]
        dir_path = pathlib.Path(dir_path).resolve()

        path_names_to_remove = [
            path_key
            for path_key, path_val in self.all_paths.items()
            if dir_path in path_val.resolve().parents
        ]
        if not rm_dir_path:
            path_names_to_remove = [p for p in path_names_to_remove if p != dir_path]
        self.rm_paths(names=path_names_to_remove)

    def get_collection_name_by_path_name(self, name: str) -> str:
        """
        Get name of the collection a path is in.

        Parameters
        ----------
        name : str
            Name of path.

        Returns
        -------
        str
            Name of collection containing the path name.
        """
        if name not in self.all_paths:
            raise ValueError("`name` was not in any of the collections.")
        for coll_name, coll_paths_dict in self._collections.items():
            if coll_paths_dict is not None and name in coll_paths_dict:
                return coll_name

    # Handle paths

    def check_paths(self):
        """
        Call path checks manually.

        These checks are automatically called after each mutation with
        setter methods, but when overwriting attributes (like check settings)
        manually, we may need to call it externally.

        May be meaningful to run after removing files and paths, as this
        does not rerun the checks.
        """
        # TODO Add better description of the checks
        self._prepare_paths()

    def _prepare_paths(self):
        # Prepare paths
        self._collections, self.all_paths = prepare_in_out_paths(
            named_collections=self._collections,
            allow_none=self.allow_none,
            allow_overwriting=self.allow_overwriting,
            allow_duplicates_in=self.allow_duplicates_in,
            disallowed_nestings=self.disallowed_nestings,
            pathlib_out=True,
            copy=True,
        )

    def mk_output_dir(
        self,
        name: str,
        messenger: Optional[Callable] = Messenger(verbose=True, indent=0, msg_fn=print),
    ):
        """
        Create non-existing output directory for a given path.

        For filepaths, it creates the directory the file is located in.

        Parameters
        ----------
        name : str
            Name of path to create output directory for.
        messenger : `utipy.Messenger` or None
            A `utipy.Messenger` instance used to print/log/... information.
            When `None`, no printing/logging is performed.
            The messenger determines the messaging function (e.g., `print`)
            and potential indentation.
        """
        path = self.get_path(name=name)
        dir_path = pathlib.Path(path).parent
        mk_dir(path=dir_path, arg_name=name, messenger=messenger)

    def mk_output_dirs(
        self,
        collection: str = None,
        messenger: Optional[Callable] = Messenger(verbose=True, indent=0, msg_fn=print),
    ):
        """
        Create non-existing output directories.

        For filepaths, it creates the directory the file is located in.

        Parameters
        ----------
        collection : str
            Name of collection to create output directories for.
                One of: ('out_dirs', 'out_files', 'tmp_files', 'mkdirs_for_tmp_dirs')
            When `None`, directories are created for all three collections.
        messenger : `utipy.Messenger` or None
            A `utipy.Messenger` instance used to print/log/... information.
            When `None`, no printing/logging is performed.
            The messenger determines the messaging function (e.g., `print`)
            and potential indentation.
        """

        # Find which collections to create output dirs for
        mkdirs_for_out_files = True
        mkdirs_for_out_dirs = True
        mkdirs_for_tmp_files = True
        mkdirs_for_tmp_dirs = True
        if collection is not None:
            if collection not in ["out_files", "out_dirs", "tmp_files", "tmp_dirs"]:
                raise ValueError(
                    f"`collection` must be one of the output path collections but was {collection}."
                )
            if collection != "out_files":
                mkdirs_for_out_files = False
            if collection != "out_dirs":
                mkdirs_for_out_dirs = False
            if collection != "tmp_files":
                mkdirs_for_tmp_files = False
            if collection != "tmp_dirs":
                mkdirs_for_tmp_dirs = False

        # Create output directories if they don't exist

        # For output directories
        if mkdirs_for_out_dirs:
            out_dirs = self.get_collection("out_dirs")
            if out_dirs is None:
                raise ValueError("`out_dirs` collection was `None`.")
            for k, v in out_dirs.items():
                mk_dir(path=v, arg_name=k, messenger=messenger)

        # For output files' directories
        if mkdirs_for_out_files:
            out_files = self.get_collection("out_files")
            if out_files is None:
                raise ValueError("`out_files` collection was `None`.")
            for k, v in out_files.items():
                # Get directory the file should be place in
                dir_path = pathlib.Path(v).parent
                mk_dir(path=dir_path, arg_name=k, messenger=messenger)

        # For tmp directories
        if mkdirs_for_tmp_dirs:
            tmp_dirs = self.get_collection("tmp_dirs")
            if tmp_dirs is None:
                raise ValueError("`tmp_dirs` collection was `None`.")
            for k, v in tmp_dirs.items():
                mk_dir(path=v, arg_name=k, messenger=messenger)

        # For tmp files' directories
        if mkdirs_for_tmp_files:
            tmp_files = self.get_collection("tmp_files")
            if tmp_files is None:
                raise ValueError("`tmp_files` collection was `None`.")
            for k, v in tmp_files.items():
                # Get directory the file should be place in
                dir_path = pathlib.Path(v).parent
                mk_dir(path=dir_path, arg_name=k, messenger=messenger)

    def rm_file(self, name: str, rm_path: bool = True, raise_on_fail: bool = True):
        """
        Remove a file from disk.

        Parameters
        ----------
        name : str
            Name of path to a file to remove from disk.
        rm_path : bool
            Whether to remove path from path collection.
            NOTE: For files that need to exist (e.g., those in the `in_files` collection),
            leaving the path after removing the file will cause downstream
            checking of the paths (see `.check_paths()`) will fail
            (as we removed the files). Those checks are called as part of
            some of the methods.
        raise_on_fail : bool
            Whether to raise an error when the path does not exist.
        """
        path = self[name]
        if path is None:
            raise ValueError(f"Path object for `{name}` was `None`.")
        if not path.is_file():
            if raise_on_fail:
                raise RuntimeError(
                    f"Path for `{name}` was not an existing file: {path}"
                )
        else:
            os.remove(str(path))

        if rm_path:
            self.rm_path(name=name)

    def rm_dir(
        self,
        name: str,
        rm_paths: bool = True,
        raise_on_fail: bool = True,
        messenger: Optional[Callable] = Messenger(verbose=True, indent=0, msg_fn=print),
    ) -> None:
        """
        Remove a directory from disk.

        Parameters
        ----------
        name : str
            Name of path to a directory to remove from disk.
        rm_paths : bool
            Whether to remove all paths that are within the
            removed directory as well as the path to the
            directory itself.
            NOTE: For files that need to exist (e.g., those in the `in_files` collection),
            leaving the path after removing the file will cause downstream
            checking of the paths (see `.check_paths()`) will fail
            (as we removed the files). Those checks are called as part of
            some of the methods.
        raise_on_fail : bool
            Whether to raise an error when the path does not exist.
        messenger : `utipy.Messenger` or None
            A `utipy.Messenger` instance used to print/log/... information.
            When `None`, no printing/logging is performed.
            The messenger determines the messaging function (e.g., `print`)
            and potential indentation.
        """
        path = self[name]
        if path is None:
            raise ValueError(f"Path object for `{name}` was `None`.")
        remove_dir(
            path=path,
            arg_name=f"{name} path",
            raise_missing=raise_on_fail,
            raise_not_dir=raise_on_fail,
            messenger=messenger,
        )
        if rm_paths:
            self.rm_paths_in_dir(dir_path=path, rm_dir_path=True)

    def rm_tmp_dirs(
        self,
        rm_paths: bool = True,
        raise_on_fail: bool = True,
        messenger: Optional[Callable] = Messenger(verbose=True, indent=0, msg_fn=print),
    ) -> None:
        """
        Remove all temporary directories from disk.

        Parameters
        ----------
        rm_paths : bool
            Whether to remove all paths that are within the
            removed directories and the paths to the directories
            themselves.
        raise_on_fail : bool
            Whether to raise an error when the path does not exist.
        messenger : `utipy.Messenger` or None
            A `utipy.Messenger` instance used to print/log/... information.
            When `None`, no printing/logging is performed.
            The messenger determines the messaging function (e.g., `print`)
            and potential indentation.
        """

        # TODO In case they are nested, we should check their existence
        # before deleting some of the directories, as that might
        # delete the existing ones
        # (I.e. find the top-level tmp dirs and remove those, and don't
        # try to remove those contained in them)

        # Delete each path in `tmp_dirs``
        for path in self.get_collection(name="tmp_dirs").keys():
            self.rm_dir(name=path, raise_on_fail=raise_on_fail, messenger=messenger)
            if rm_paths:
                self.rm_paths_in_dir(dir_path=path, rm_dir_path=True)

    def mv_file(
        self,
        name: str,
        new_path: Union[str, pathlib.PurePath],
        update_path: bool = True,
    ) -> None:
        """
        Move a file to a new path.
        Optionally update the path for `name` in the collection.

        When updating the path after moving the file, it re-checks
        the paths. See `.check_paths()` for details.
        """
        coll_name = self.get_collection_name_by_path_name(name=name)
        self[name].rename(new_path)
        if update_path:
            self._collections[coll_name][name] = new_path
            self._prepare_paths()

    def update(self, other: object):
        """
        Update with paths from another `IOPaths` collection.

        Note: When a collection in `other` is `None`, we
        keep the original collection.

        Parameters
        ----------
        other : `IOPaths` instance
            Another `IOPaths` instance with paths.
            Each sub collection is dict-updated one at a time.
        """
        assert isinstance(other, IOPaths)
        for coll_name in IOPaths.COLLECTION_NAMES:
            other_paths = other.get_collection(name=coll_name)
            if other_paths is None:
                continue
            self._update_collection(paths=other_paths, collection=coll_name)

        # Ensure consistency
        self._prepare_paths()

    def difference(self, other: object):
        """
        Find the combinations of keys and paths in the collections of this object
        that are not in the collections of the `other` object.
        Creates a new `IOPaths` object with the settings from this object.

        Parameters
        ----------
        other : `IOPaths` instance
            Another `IOPaths` instance with paths.
            Each sub collection is dict-updated one at a time.

        Returns
        -------
        `IOPaths` instance
            A new `IOPaths` instance with the keys and paths in
            the collections of this object that are not in the collections
            of the `other` object.
        """
        assert isinstance(other, IOPaths)
        diff_dicts = {}
        for coll_name in IOPaths.COLLECTION_NAMES:
            this_coll = self.get_collection(name=coll_name)
            if this_coll is None:
                diff_dicts[coll_name] = None
                continue
            other_coll = other.get_collection(name=coll_name)
            if other_coll is None:
                diff_dicts[coll_name] = this_coll
                continue
            this_coll = set(this_coll.items())
            other_coll = set(other_coll.items())
            diff_dicts[coll_name] = dict(this_coll.difference(other_coll))
        return self._new(
            in_files=diff_dicts["in_files"],
            in_dirs=diff_dicts["in_dirs"],
            out_files=diff_dicts["out_files"],
            out_dirs=diff_dicts["out_dirs"],
            tmp_files=diff_dicts["tmp_files"],
            tmp_dirs=diff_dicts["tmp_dirs"],
        )

    def __eq__(self, other: object) -> bool:
        """
        Check equality of this `IOPaths` instance and another.

        Checks that all sub collections are the same.
        No other attributes are considered.

        Parameters
        ----------
        other : `IOPaths` instance
            Another `IOPaths` instance with paths.
        """
        assert isinstance(other, IOPaths)
        if self.npaths != other.npaths:
            return False
        for coll_name in IOPaths.COLLECTION_NAMES:
            this_coll = self.get_collection(name=coll_name)
            other_coll = other.get_collection(name=coll_name)
            if sum([other_coll is None, this_coll is None]) == 1:
                return False
            if this_coll != other_coll:
                return False
        return True

    def _new(
        self,
        in_files: dict = None,
        in_dirs: dict = None,
        out_files: dict = None,
        out_dirs: dict = None,
        tmp_files: dict = None,
        tmp_dirs: dict = None,
    ):
        """
        Create new `IOPaths` object with new collections but keeping the rest of the settings.

        Parameters
        ----------
        in_files : dict
            Dict with named paths to input files, mapping `(argument) name -> input filepath`.
            A path can also be "-" when streaming input.
            This value will remain a string and will not be checked
            for duplication, existence, etc.
        in_dirs : dict
            Dict with named paths to input directories, mapping `(argument) name -> input directory path`.
        out_files : dict
            Dict with named paths to output files, mapping `(argument) name -> output filepath`.
        out_dirs : dict
            Dict with named paths to output directories, mapping `(argument) name -> output directory path`.
        tmp_files : dict
            Dict with named paths to temporary files, mapping `(argument) name -> temporary filepath`.
        tmp_dirs : dict
            Dict with named paths to temporary directories, mapping `(argument) name -> temporary directory path`.

        Returns
        -------
        `IOPaths` instance
        """
        return IOPaths(
            in_files=in_files,
            in_dirs=in_dirs,
            out_files=out_files,
            out_dirs=out_dirs,
            tmp_files=tmp_files,
            tmp_dirs=tmp_dirs,
            allow_none=self.allow_none,
            allow_overwriting=self.allow_overwriting,
            allow_duplicates_in=self.allow_duplicates_in,
            print_note=self.print_note,
        )

    def _check_collection_name(self, name: str):
        if name not in IOPaths.COLLECTION_NAMES:
            raise ValueError(f"Collection name was unknown: {name}.")

    def _format_path_out(self, path, as_str: bool):
        if as_str:
            path = str(path)
        return path

    def __str__(self) -> str:
        """
        Convert object to a string with the paths for each sub collection.
        """
        max_paths_per_coll = 10
        lines = []
        lines.append("Input and output paths")
        for coll_name in IOPaths.COLLECTION_NAMES:
            collection = self.get_collection(name=coll_name)
            if collection is None:
                lines.append(f"  {coll_name} (0)")
            else:
                lines.append(f"  {coll_name} ({len(collection)}):")
                sorted_keys = sorted(collection.keys())
                for i, key in enumerate(sorted_keys):
                    if i < max_paths_per_coll:
                        lines.append(f"    {key} -> {collection[key]}")
                    else:
                        lines.append(f"    ...")
                        break
        if self.print_note:
            lines.append(f"  Note: {self.print_note}")
        return "\n".join(lines)



"""
@author: ludvigolsen
"""

import pandas as pd
from numbers import Number
from typing import Optional
from utipy.groups.methods.l_sizes import _l_sizes
from utipy.groups.methods.n_dist import _n_dist


def group(
    data: pd.DataFrame,
    n: Number,
    col: Optional[str] = None,
    method: str = 'n_dist',
    copy: bool = True
) -> pd.DataFrame:
    """
    Add a grouping factor to dataframe.

    """
    if copy:
        data = data.copy()

    # If no column is given
    # We use the first column in data
    if col is None:

        # Get first column of data
        column = data.ix[:, 0]

        # Create grouping factor
        if method == 'n_dist':
            all_group_ids = _n_dist(column, n, randomize=True)
        elif method == 'l_sizes':
            all_group_ids = _l_sizes(column, n, randomize=True)

    # Else we use the given column
    else:

        # Create grouping factor
        if method == 'n_dist':
            all_group_ids = _n_dist(data[col], n, randomize=True)
        elif method == 'l_sizes':
            all_group_ids = _l_sizes(data[col], n, randomize=True)

    # Add grouping factor to data
    data['group'] = all_group_ids

    return data



"""
@author: ludvigolsen
"""

from typing import Any
import numpy as np
import pandas as pd


def check_instance(data: Any) -> str:
    """
    Checks if data is either:
        list
        tuple
        np.ndarray
        pd.Series
        pd.DataFrame

    """
    if isinstance(data, list):
        return 'list'
    elif isinstance(data, tuple):
        return 'tuple'
    elif type(data).__module__ == np.__name__:
        return 'np.ndarray'
    elif isinstance(data, pd.Series):
        return 'pd.Series'
    elif isinstance(data, pd.DataFrame):
        return 'pd.DataFrame'
    else:
        raise TypeError("Doesn't recognize instance type")




import random
from string import ascii_letters, digits
from typing import Optional


def random_alphanumeric(size: int, seed: Optional[int] = None) -> str:
    """
    Create a string with random alphanumeric characters.

    Uses it's own `random.Random()` generator why it is
    not affected by external seeds.

    Parameters
    ----------
    size : int 
        Number of characters in the random string.
    seed : int or `None`
        A random seed to use.

    Returns
    -------
    str
        A string of length `size` with random alphanumeric characters.
    """
    # Create a unique random generator
    # which is not affected by seeds elsewhere
    randomgen = random.Random()

    # Set random seed when given
    if seed is not None:
        randomgen.seed(seed)

    # Collect all alphanumeric characters
    char_choices = ascii_letters + digits

    # Pick `size` characters and join to a string
    return ''.join(randomgen.choices(char_choices, k=size))



from datetime import datetime
from contextlib import contextmanager
from typing import Any, Callable, Optional, Union, List


class Messenger:
    def __init__(
        self,
        verbose: bool = True,
        msg_fn: Callable = print,
        indent: Union[int, None] = 0,
        **kwargs: Any,
    ) -> None:
        """
        Wraps a messaging function (print/log.info/...) to simplify messenging.

        The `Messenger` object has specified defaults for whether to perform the messaging (`verbose`),
        which function to use (`msg_fn`) and how much to indent messages (`indent`).

        This removes `if` statements before each print/log/... call for checking
        whether we should perform the messaging or not (`verbose`).


        Parameters
        ----------
        verbose : bool
            Default value for whether to perform the messaging.
        msg_fn : callable
            Function for performing the messaging.
            E.g., `print` or `log.info`.
        indent : int
            Default value for number of whitespaces to indent the message.
        kwargs : keyword arguments
            Named arguments to pass to the `msg_fn` by default.
            The arguments can be overwritten for single calls via the `kwargs` arguments in `__call__()`.
            I.e. by passing them again with different values during the call.


        Examples
        --------

        Initialize print messenger. By default, the `print` function is used. Set an indentation to use.

        >>> printer = Messenger(verbose=True, indent=4)

        Print a string. Notice the added indentation.

        >>> printer("anything printable")
        "    anything printable"

        We can set a specific indentation for a call.

        >>> printer("overriding indentation for call", indent=0)
        "overwriting indentation for call"

        We can pass keyword arguments to the messaging function (here `print()`).

        >>> printer("a string", end=' is all you need')
        "    a string is all you need"

        Create a logging messenger with the logging library.
        Set level to INFO so we see the output.

        >>> import logging
        >>> logging.basicConfig()
        >>> _logger = logging.getLogger("Mr.Logger")
        >>> _logger.setLevel(logging.INFO)
        >>> logger = Messenger(verbose=True, indent=0, msg_fn=_logger.info)

        Log something.

        >>> logger("anything printable", "- even multiple arguments")
        "INFO  Mr.Logger:script.py:xz  anything printable - even multiple arguments"
        """
        # Check types
        assert isinstance(verbose, bool)
        assert callable(msg_fn)
        assert isinstance(indent, int)

        self._verbose = verbose
        self.msg_fn = msg_fn
        self._indent = indent
        self.kwargs = kwargs

    def set_verbose(self, verbose: bool) -> None:
        """
        Set the `verbose` state of the messenger.

        Parameters
        ----------
        verbose : bool
            Default value for whether to perform the messaging.
        """
        self._verbose = verbose

    @property
    def verbose(self) -> bool:
        """
        Get the `verbose` state of the messenger.

        Returns
        -------
        bool
            The default value for whether to perform the messaging.
        """
        return self._verbose

    def set_indent(self, indent: int) -> None:
        """
        Set the indentation of the messenger.

        Parameters
        ----------
        indent : int
            Default value for number of whitespaces to indent the message.
        """
        if indent < 0:
            raise ValueError("`indent` was negative.")
        self._indent = indent

    @property
    def indent(self) -> int:
        """
        Get the indentation setting in the messenger.

        Returns
        -------
        int
            Default number of whitespaces that messages are indented.
        """
        return self._indent

    def __call__(
        self,
        *objects: Any,
        verbose: Union[None, bool] = None,
        indent: Union[None, int] = None,
        add_indent: Union[None, int] = None,
        sep: str = " ",
        add_msg_fn: Optional[Callable] = None,
        **kwargs: Any,
    ) -> None:
        """
        Perform messaging using `self.msg_fn`.

        Parameters
        ----------
        objects : objects
            Objects to message. Anything printable (i.e. with a `__str__` method).
        verbose : bool
            Whether to perform the messaging for this specific call.
        indent : int
            Number of whitespaces to indent the message in this specific call.
        add_indent : int
            Number of whitespaces to add to the current indent state
            for the message in this specific call.
        sep : str
            String used to separate `objects`.
        add_msg_fn : Callable or `None`
            Additional messaging function to use.
            E.g., `warnings.warn` for throwing a warning as well.
        kwargs : keyword arguments
            Named arguments for the messaging function.
        """
        if verbose is None:
            verbose = self._verbose
        if add_indent is not None:
            if indent is not None:
                raise ValueError(
                    "`indent` and `add_indent` were both specified. "
                    "Please only specify one (or none) of them."
                )
            indent = self._indent + add_indent
        if indent is None:
            indent = self._indent

        # Get kwargs for current call
        call_kwargs = self.kwargs.copy()
        call_kwargs.update(kwargs)

        # Set messenging functions
        msg_fn = [self.msg_fn]
        if add_msg_fn is not None:
            msg_fn.append(add_msg_fn)

        msg_if(
            *objects,
            verbose=verbose,
            indent=indent,
            msg_fn=msg_fn,
            sep=sep,
            **call_kwargs,
        )

    def __add__(self, indent: int):
        """
        Add to the current indentation level.

        Parameters
        ----------
        indent : int
            The number of additional positions to indent message.

        Returns
        -------
        `self`
        """
        assert isinstance(indent, int)
        if self._indent + indent < 0:
            raise ValueError(
                f"Subtracting {abs(indent)} positions would give a "
                "negative indentation level. Can maximally subtract "
                f"{self._indent} positions."
            )
        self._indent += indent
        return self

    def __sub__(self, indent: int):
        """
        Subtract from the current indentation level.

        Parameters
        ----------
        indent : int
            The number of fewer positions to indent message.

        Returns
        -------
        `self`
        """
        assert isinstance(indent, int)
        return self.__add__(indent=indent)

    @contextmanager
    def indentation(
        self, indent: Optional[int] = None, add_indent: Optional[int] = None
    ) -> None:
        """
        Function to use in `with` statement. Temporarily changes the indentation.
        The original indentation is restored upon exiting the `with` context.

        Can be used in a nested fashion.

        Parameters
        ----------
        indent : int, optional
            A specific indentation to use within the context.
            Exactly one of `indent` and `add_indent` should be specified.
        add_indent : int, optional
            A relative change in indentation based on the current indentation.
            Exactly one of `indent` and `add_indent` should be specified.

        Yields
        ------
        `None` once.

        Examples
        --------

        Initialize a messenger with `2` indentation spaces.

        >>> msg = Messenger(indent=2)

        Use `4` indentation spaces in a given context.
        After exiting the `with` context, the indentation
        is restored to `2` spaces.

        >>> with msg.indentation(indent=4):
        >>>     msg("something")
        "    something"

        Add `4` spaces of indentation to the existing indentation of `2` spaces.

        >>> with msg.indentation(add_indent=4):
        >>>     msg("something")
        "      something"

        Check that indentation is back to `2` spaces.

        >>> msg("something")
        "  something"
        """
        if sum([indent is not None, add_indent is not None]) != 1:
            raise ValueError(
                "Exactly one of {'indent', 'add_indent'} should be specified."
            )

        # Store original indentation, allowing us to restore it
        original_indent = self.indent
        if indent is not None:
            self.set_indent(indent=indent)
        if add_indent is not None:
            self.__add__(indent=add_indent)

        try:
            yield None
        finally:
            self.set_indent(indent=original_indent)

    def now(
        self,
        message: str = "Current time:",
        time_format: str = "%a, %d %b, %Y %H:%M:%S",
        verbose: Union[None, bool] = None,
        indent: Union[None, int] = None,
        sep: str = " ",
        **kwargs: Any,
    ) -> None:
        """
        Message the current date and time.

        Parameters
        ----------
        message : str
            Prefix string.
        time_format : str
            Time formatting string for the `datetime` `.strftime()` method.
            Default leads to "Thu, 26 Jan, 2023 10:53:48"-style strings.
        verbose : bool
            Whether to perform the messaging for this specific call.
        indent : int
            Number of whitespaces to indent the message in this specific call.
        sep : str
            String used to separate `objects`.
        kwargs : keyword arguments
            Named arguments for the messaging function.
        """
        now = datetime.now()
        now = now.strftime(time_format)
        self(message, now, verbose=verbose, indent=indent, sep=sep, **kwargs)


def msg_if(
    *objects: Any,
    verbose: bool,
    indent: int = 0,
    msg_fn: Union[Callable, List[Callable]] = print,
    sep: str = " ",
    **kwargs,
) -> None:
    """
    Message (print/log/..) the given `objects` arguments when `verbose` is enabled.

    Parameters
    ----------
    objects : Any
        Objects to message. Anything printable (i.e. with a `__str__` method).
    verbose : bool
        Whether to perform the messaging.
    indent : int
        Number of whitespaces to indent the message.
    msg_fn : callable or list of callables
        Function(s) for performing the messaging.
        E.g., `print` or `log.info`.
    sep : str
            String used to separate `objects`.
    kwargs : keyword arguments
        Named arguments for the messaging function.
    """
    assert indent >= 0
    if verbose:
        subtract_1 = len(objects) > 0
        indent_str = "".join([" " for _ in range(max(0, indent - int(subtract_1)))])
        if not isinstance(msg_fn, list):
            msg_fn = [msg_fn]
        for fn in msg_fn:
            fn(
                _objects_to_string(
                    indent_str,
                    *objects,
                    sep=sep,
                ),
                **kwargs,
            )


def _objects_to_string(*args: Any, sep: str = " "):
    try:
        obj_as_strings = [str(a) for a in args]
    except Exception as e:
        raise RuntimeError(f"Failed to convert argument to string: {e}")
    return f"{sep}".join(obj_as_strings)


def check_messenger(messenger: Optional[Callable]) -> Messenger:
    """
    Check that `messenger` is a `utipy.Messenger` object or `None`.
    In the latter case a `utipy.Messenger` with `verbose=False` is returned.

    Parameters
    ----------
    messenger : `utipy.Messenger` or None
        A Messenger instance to check.
        Or `None`, in which case a `utipy.Messenger` with `verbose=False` is returned.

    Returns
    -------
    `utipy.Messenger`
    """
    # Check the messenger function
    if messenger is None:
        messenger = Messenger(verbose=False)
    else:
        assert isinstance(messenger, Messenger)
    return messenger



import time
from typing import Any, Union
import numpy as np
import pandas as pd

from .format_time import format_time_hhmmss


class Timestamps:
    def __init__(self) -> None:
        """
        Container for storing timestamps
        and calculating the difference between two
        timestamps (e.g., the latest 2).
        """
        self.timestamps = []
        self.name_to_idx = {}
        self.idx_to_name = {}

    def __len__(self) -> int:
        """
        Number of stored timestamps.
        """
        return len(self.timestamps)

    def __eq__(self, other: object) -> bool:
        """
        Equality test.

        Checks whether the list of timestamps and the dict
        mapping `stamp name -> index` are equal.

        Parameters
        ----------
        other : `Timestamps`
            Another `Timestamps` collection.

        Returns
        -------
        bool
            Whether the two `Timestamps` collections are equal.
        """
        return (
            self.timestamps == other.timestamps
            and self.name_to_idx == other.name_to_idx
        )

    def __getitem__(self, idx_or_name: Union[str, int]) -> int:
        """
        Get numeric timestamp (as recorded with `time.time()`)
        by indexing (via idx or name) in square brackets.

        Parameters
        ----------
        idx_or_name : int or str
            Index or name of timestamp.
        """
        assert isinstance(idx_or_name, (str, int))
        key = "idx" if isinstance(idx_or_name, int) else "name"
        return self.get_stamp(**{key: idx_or_name}, as_str=False)

    def __str__(self) -> str:
        string = "Timestamps:\n\n"
        return string + self.to_data_frame().to_string(max_rows=30) + "\n"

    def _stamp(self) -> None:
        """
        Add current time to list of timestamps.
        """
        t = time.time()
        self.timestamps.append(t)

    def stamp(self, name: str = None) -> None:
        """
        Add current time to list of timestamps.

        *Optionally* save the timestamp index with a name in the
        `name_to_idx` dictionary to allow easy extraction later.
        E.g., use to get the time difference between
        two larger blocks of code with in-between timestamps.

        Parameters
        ----------
        name : str
            (Optional) Unique name to store the index of the timestamp with.
        """
        self._stamp()
        if name is not None:
            if name in self.name_to_idx:
                raise ValueError("`name` was already used. Use a unique name.")
            idx = len(self) - 1
            self.name_to_idx[name] = idx
            self.idx_to_name[idx] = name

    def get_stamp(
        self,
        idx: Union[int, None] = None,
        name: Union[str, None] = None,
        as_str: bool = True,
    ) -> Union[int, str]:
        """
        Get specific timestamp from either the index or name it was recorded under.

        Note: The raw list of timestamps are also available as `.timestamps`
        while the `name->index` dict is available as
        `.name_to_idx`.

        Parameters
        ----------
        idx : int
            Index of the timestamp to get.
        name : str
            Name of the timestamp.
        as_str : bool
            Whether to format the difference as a string.

        Returns
        -------
        int or str
            Timestamp made with `time.time()`.
            Optionally formatted as a string with hh:mm:ss.
        """
        if sum([idx is not None, name is not None]) != 1:
            raise ValueError("Exactly one of `idx` and `name` should be specified.")
        if idx is not None:
            t = self.timestamps[idx]
        else:
            if name not in self.name_to_idx:
                raise KeyError(f"`name` '{name}' was not found.")
            t = self.timestamps[self.get_stamp_idx(name=name)]
        if as_str:
            t = format_time_hhmmss(t)
        return t

    def get_stamp_idx(self, name: str) -> int:
        """
        Get timestamp index from name of timestamp.

        Parameters
        ----------
        name : str
            Name used to store index of timestamp with.

        Returns
        -------
        int
            Index of timestamp stored with `name`.
        """
        if name not in self.name_to_idx:
            raise KeyError(f"`name` '{name}' was not found.")
        return self.name_to_idx[name]

    def get_stamp_name(self, idx: int) -> str:
        """
        Get timestamp name from index of timestamp.

        Parameters
        ----------
        idx : int
            Index of timestamp to get name for.

        Returns
        -------
        str or None
            Name of timestamp at the given index.
            When no name was recorded, `None` is returned.
        """
        if np.abs(idx) > len(self) - 1:
            raise ValueError(
                f"`idx` was out of bounds: '{idx}'. Currently stores {len(self)} timestamps."
            )
        return self.idx_to_name.get(idx, None)

    def to_data_frame(self):
        """
        Get times as `pandas.DataFrame` with columns [`Name`, `Time`].

        Returns
        -------
        `pandas.DataFrame`
            Data frame with names and times in the recorded order.
        """
        names = [""] * len(self)
        for idx, name in self.idx_to_name.items():
            names[idx] = name
        times = self.timestamps.copy()
        times_from_start = [t - times[0] for t in times]
        times_from_start = [format_time_hhmmss(t) for t in times_from_start]
        return pd.DataFrame(
            {"Name": names, "Time Raw": times, "Time From Start": times_from_start}
        )

    def to_csv(self, *args: Any, **kwargs: Any) -> None:
        """
        Write timestamps to a csv file.

        Converts a copy of the collection to a `pandas.DataFrame`
        and saves it with `pandas.DataFrame.to_csv()`.

        Parameters
        ----------
        args
            positional arguments for `pandas.DataFrame.to_csv()`.
        kwargs
            keyword arguments for `pandas.DataFrame.to_csv()`.
        """
        df = self.to_data_frame()
        df.to_csv(*args, **kwargs)

    def took(
        self,
        start: Union[int, str] = -2,
        end: Union[int, str] = -1,
        as_str: bool = True,
        raise_negative: bool = True,
    ) -> Union[int, str]:
        """
        Get the difference between two timestamps.
        By default, the two latest timestamps are used.

        Parameters
        ----------
        start : int or str
            Either:
                1) The index of the starting timestamp.
                2) The name of the starting timestamp.
        end_idx : int or str
            Either:
                1) The index of the end timestamp.
                2) The name of the end timestamp.
        as_str : bool
            Whether to format the difference as a string.
        raise_negative : bool
            Whether to raise an error when the time difference
            between the two stamps are negative.
            In thus case, `end` came before `start`.
            When `False`, a negative number is returned.

        Returns
        -------
        int or str
            Difference in time between two given timestamps,
            either as a number or a formatted string.
        """
        start_time = self.get_stamp(
            idx=None if not isinstance(start, int) else start,
            name=None if not isinstance(start, str) else start,
            as_str=False,
        )
        end_time = self.get_stamp(
            idx=None if not isinstance(end, int) else end,
            name=None if not isinstance(end, str) else end,
            as_str=False,
        )
        diff = end_time - start_time
        if diff < 0 and raise_negative:
            raise ValueError(
                (
                    "Difference between timestamps was negative. "
                    "`start` should correspond to an earlier timestamp than `end` "
                    "(or disable `raise_negative`)."
                )
            )
        if as_str:
            return format_time_hhmmss(diff)
        return diff

    def get_total_time(self, as_str: str = True) -> Union[int, str]:
        """
        Get the time difference between the first and last timestamps.

        Parameters
        ----------
        as_str : bool
            Whether to format the difference as a string.

        Returns
        -------
        int
            Difference in time between first and last timestamp,
            either as a number or a formatted string.
        """
        return self.took(start=0, end=-1, as_str=as_str)

    def update(self, other: object):
        """
        Update existing `Timestamps` collection with another `Timestamps` collection.
        Combines the list of timestamps (sorted by time) and updates the name->idx and idx->name maps,
        which is likely to change the indices.

        When both collections have the same key in the `.name_to_idx` and `.idx_to_name` dicts,
        the member from `other` is used.

        Parameters
        ----------
        other : `Timestamps` object
            Another `Timestamps` collection to combine with the current collection.
        """
        self.merge(other=other, suffix_identical_names=False)

    def merge(self, other: object, suffix_identical_names: bool = True):
        """
        Merge this `Timestamps` collection with another `Timestamps` collection.
        Combines the list of timestamps (sorted by time) and updates the name->idx and idx->name maps,
        which is likely to change the indices.

        By default, clashing names are suffixed with "_0" (this) and "_1" (other).
        If this does not lead to unique names, the suffix increases by one ("_2", "_3", "_4", ...)
        until it does.

        Parameters
        ----------
        other : `Timestamps` object
            Another `Timestamps` collection to combine with the current collection.
        suffix_identical_names : bool
            Whether to add a suffix ("_0", "_1", etc.) to clashing names
            with an increasing count until names are unique.
            When `False`, the dict members in `.name_to_idx` and `.idx_to_name`
            from `other` is used.
        """
        # Add the two timestamps lists together
        # but as tuples with indices and an identifier
        # of which collection it came from
        combined_timestamps = _list_to_enumerated_tuple(
            l=self.timestamps, identifier="this"
        ) + _list_to_enumerated_tuple(l=other.timestamps, identifier="other")

        # Sort by ascending time
        combined_timestamps = sorted(combined_timestamps, key=lambda x: x[0])

        # Update name->index maps
        colls = {"this": self, "other": other}
        for new_idx, (_, old_idx, coll_id) in enumerate(combined_timestamps):
            coll = colls[coll_id]
            name = coll.idx_to_name.get(old_idx, None)
            if name is not None:
                coll.name_to_idx[name] = new_idx

        # Handle clashing names
        # Either by suffixing or overwriting
        if suffix_identical_names:
            # Find the clashing names
            all_names = list(self.name_to_idx.keys()) + list(other.name_to_idx.keys())
            duplicate_names = set(self.name_to_idx.keys()).intersection(
                other.name_to_idx.keys()
            )
            if duplicate_names:
                for coll in [self, other]:
                    for name in duplicate_names:
                        new_name = _create_unique_name(name, all_names)
                        all_names += [new_name]
                        coll.name_to_idx[new_name] = coll.name_to_idx.pop(name)

        # Update the dict
        self.name_to_idx.update(other.name_to_idx)

        # Create idx_to_name dict
        self.idx_to_name = {v: k for k, v in self.name_to_idx.items()}

        # Assign the combined timestamps
        self.timestamps = [x[0] for x in combined_timestamps]


def _list_to_enumerated_tuple(l, identifier):
    def make_tuple(t, i, identifier):
        if identifier is None:
            return (t, i)
        return (t, i, identifier)

    return [make_tuple(t, i, identifier) for i, t in enumerate(l)]


def _create_unique_name(name, l):
    # Add underscore
    name += "_"
    counter = 0
    while name + str(counter) in l:
        counter += 1
    return name + str(counter)



#!/usr/bin/env
# -*- coding: utf-8 -*-
"""
@author: ludvigolsen
"""

# Import operator module for dynamically passing operators
from typing import Union
from numbers import Number
import pandas as pd
import operator
import numpy as np
from utipy.utils.convert_to_type import convert_to_type


def makes_up(
    x: Union[pd.Series, np.ndarray, list],
    value: Union[str, Number],
    thresh: float,
    direction: str = '>',
    missing_error: bool = False
):
    """
    Checks the percent-wise appearance of a specific value (or any value) in a Series
    against a threshold and a given direction.

    Asks:
        'Does [value] make up [direction] than [thresh] percent of `x`?'
    E.g.:
        'Does 0 make up more than 90 percent of `x`?'


    Parameters
    ----------
    x : pd.Series, np.ndarray, list
        The values to check. Array-like objects will be converted to pd.Series internally.
    value : str / int / float
        The value to match.
            Regular value,  
            'any',  
            'NaN',  
            'inf'  
    thresh : float
        Threshold.
        Percentage between 0-1.
    direction : str
        Operator sign for comparison.
            '>', '<', '>=', '<=', '=='.
    missing_error : bool
        Raise error if value is not in `x`.


    Returns
    -------
    bool


    Examples
    --------
    
    Does '0' make up less than 30% of `x`?
    >>>  makes_up(
    ...     x, 
    ...     value = '0', 
    ...     thresh = 0.3, 
    ...     direction = '<'
    ... )

    Any NaNs in `x`? 

    >>> makes_up(
    ...     x, 
    ...     value = 'NaN', 
    ...     thresh = 0, 
    ...     direction = '>'
    ... )

    Does `x` only contain 1 unique value?
    I.e. the same value in 100% of the rows.
    >>>  makes_up(
    ...     x, 
    ...     value = 'any',
    ...     thresh = 1, 
    ...     direction = '=='
    ...  )

    """

    # Make sure directioon is given as '>', '<', '<=', '>=', or '=='
    if not direction in ['>', '<', '<=', '>=', '==']:
        raise ValueError(
            '`direction` can only be given as \'>\',\'<\',\'>=\',\'<=\', or \'==\'.')

    # Create dictionary for calling operator
    # Uses module: operator
    ops = {'>': operator.gt,
           '<': operator.lt,
           '>=': operator.ge,
           '<=': operator.le,
           '==': operator.eq}

    # Make sure thresh is given as a number between 0 and 1
    if not (thresh >= 0 and thresh <= 1):
        raise ValueError(
            '`thresh` had an incorrect format. '
            'Pass as a percentage (between 0-1).')

    # Check if x is a pd.Series
    # and convert if necessary
    x = convert_to_type(x, 'pd.Series')

    # If it is an object, we can only recognize NaN and inf as strings
    # and fall back to checking it as a string
    if x.dtype in [np.dtype('object')] and value in ['NaN', 'inf', 'any']:
        # When converted to string,
        # np.nan becomes 'nan'
        if value == 'NaN':
            value = 'nan'

        # Make sure that object elements are considered strings
        x = x.apply(str)

        # First check if value is 'any'
        if value == 'any':
            n_value = x.value_counts(dropna=False).max()

        # First we check if value is in x at all
        elif not (value in set(x)):
            # If asked to raise error if value is not found
            # raise error
            if missing_error:
                raise ValueError('`value` not found in `x`')
            # Else, it was found 0 times in series
            else:
                n_value = 0
        else:
            # If it IS in `x``,
            # count how many times
            n_value = x.value_counts(dropna=False)[value]

    elif value == 'NaN':
        n_value = x.isnull().sum()

    elif value == 'inf':
        try:
            n_value = np.isinf(x).sum()
        except:
            raise ValueError(
                "`value` ('inf') could not be searched for in `x`.")

    elif value == 'any':
        n_value = x.value_counts(dropna=False).max()

        # If n_value is NaN we want to convert it
        # to a number instead. So the number of
        # NaNs.
        if np.isnan(n_value):
            n_value = x.isnull().sum()

    # Check if value is not in `x`
    elif value not in set(x):
        # If asked to raise error if value is not found
        if missing_error:
            raise ValueError('value not found in `x`')
        # Else, it was found 0 times in series
        else:
            n_value = 0
    # If value in `x``
    else:
        # Get how many times value is in col
        n_value = x.value_counts()[value]

    # Get total number of values in col
    total_values = len(x)

    # Check with the given direction (operator) if n_value
    # exceeds the threshold
    return ops[direction]((n_value / float(total_values)), thresh)



"""
@author: ludvigolsen
"""

from typing import Union
import numpy as np
import pandas as pd
from utipy.utils.check_instance import check_instance
from utipy.utils.convert_to_type import convert_to_type

def windowed_reverse(x: Union[list, np.ndarray, pd.Series], wsize:int = 2) -> Union[list, np.ndarray, pd.Series]:
    
    """Reverse / flip windows of array

    Parameters
    ----------
    x : list, np.ndarray, pd.Series
        The array to reverse
    wsize: int
    	Window size

    Returns
    -------
    list, np.ndarray, pd.Series
    	Window reversed array with the type of the original

    Examples
    --------
	x = 1,2,3,4,5,6; wsize = 2 returns 2,1,4,3,6,5
	x = 1,2,3,4,5,6; wsize = 3 returns 3,2,1,6,5,4
    """

    # Get instance type (np.ndarray, list, pd.Series)
    instance_type = check_instance(x)
    
    # Window and reverse
    reversed_windows = [x[pos:pos+wsize][::-1] for pos in range(len(x))[::wsize]]
    
    # Flatten windows
    flattened_array = np.concatenate(reversed_windows)
    
    # Convert to original type (np.ndarray, list, pd.Series)
    return convert_to_type(flattened_array, instance_type)
    


# outer __init__.py

from .pandas.drop import drop
from .pandas.makes_up import makes_up
from .pandas.resemble import resemble
from .pandas.distort import distort
from .pandas.subset_by_levels import subset_by_levels
from .pandas.polynomializer import polynomializer
from .pandas.move_column_inplace import move_column_inplace

from .groups.fold import fold
from .groups.group_uniques import group_uniques
from .groups.group import group
from .groups.partition import partition

from .array.blend import blend
from .array.windowed_reverse import windowed_reverse
from .array.window import window
from .array.nan_stats import nan_stats, print_nan_stats

from .time.timestamps import Timestamps
from .time.timer import StepTimer

from .path.iopaths import IOPaths
from .path.mk_rm_dir import mk_dir, rm_dir

from .string.letter_strings import letter_strings
from .string.random_strings import random_alphanumeric

from .utils.messenger import Messenger, check_messenger, msg_if


def get_version():
    import importlib.metadata

    return importlib.metadata.version("utipy")


__version__ = get_version()



"""
@author: ludvigolsen
"""

from typing import Union
from utipy.utils import _extended_describe
import numpy as np
import pandas as pd
from random import shuffle


def resemble(
    x: Union[pd.Series, np.ndarray, list],
    distribution: str = 'uniform'
) -> pd.Series:
    """
    Generate pandas.Series that resembles `x`.

    Generates data that resembles the original data by using descriptors
    of `x` to create and sample from a specified distribution.


    Parameters
    ----------
    x : pandas.Series, numpy.ndarray, list
        The values to resemble.
    distribution : str
        Distribution to sample from. One of:
            'uniform'
                Between min. and max.
            'gaussian'
                From mean and std.
            'robust gaussian'
                From median and IQR.
            'shuffle'
                Shuffles original data.


    Returns
    -------
    pd.Series
    """

    # TODO Do a tryCatch on describe to
    # see if it's a character vector or other
    # type that it can't describe.
    # Alternatively just check if input is int, double or float?
    # Get descriptors

    # TODO Check for NaNs

    # Get description of `x`
    desc = _extended_describe(x)

    if distribution == 'uniform':
        generated = np.random.uniform(
            low=desc['min'],
            high=desc['max'],
            size=int(desc['count'])
        )

    elif distribution == 'gaussian':
        generated = np.random.normal(
            loc=desc['mean'],
            scale=desc['std'],
            size=int(desc['count'])
        )

    elif distribution == 'robust gaussian':
        generated = np.random.normal(
            loc=desc['median'],
            scale=desc['IQR'],
            size=int(desc['count'])
        )

    elif distribution == 'poisson':
        # TODO Check up on this one. Should there be a min/max?
        generated = np.random.poisson(
            lam=desc['max'],
            size=int(desc['count'])
        )

    elif distribution == 'shuffle':
        generated = shuffle(x)

    # Change back to original dtype
    generated = generated.astype(desc['dtype'])

    return generated




from numbers import Number
import numpy as np

# TODO cythonize
# TODO add tests

def format_time_hhmmss(t: int) -> str:
    """
    Format a count of seconds into hh:mm:ss where hh is allowed to be >24 
    (i.e. hhh, hhhh or even MORE 'h's!).

    Parameters
    ----------
    t : int
        Timepoint as found with `time.time()`.

    Returns
    -------
    str
        The timepoint formatted as a hh:mm:ss string.
    """
    # Handle negative numbers
    is_negative = t < 0
    if is_negative:
        t *= -1

    # Hours
    hours = np.floor(t / 60 / 60)
    hours_seconds = hours * 60 * 60

    # Minutes
    mins = np.floor((t - hours_seconds) / 60)
    mins_seconds = mins * 60

    # Seconds
    secs = (t - hours_seconds - mins_seconds)

    neg_string = "-" if is_negative else ""
    return f"{neg_string}{_format_part(hours)}:{_format_part(mins)}:{_format_part(secs)}"


def _format_part(t: Number) -> str:
    """
    Format integer-like number as 2-digit string (either hh, mm, or ss).

    Parameters
    ----------
    t : Number
        A 1- or 2-digit number (after conversion to integer).

    Returns
    -------
    str
        A two-character string. From "00" -> "99".
    """
    t = int(t)
    if t == 0:
        return "00"
    if t < 10:
        return "0" + str(t)
    return str(t)



"""
@author: ludvigolsen
"""

import numpy as np
from random import shuffle


def _l_sizes(v, p, randomize=False, rounding='floor'):
    """
    Creates grouping factor from group size(s)

    p is given as percentage (0-1)

    rounding: Floors percentages by default
    Alternative setting: 'round' or 'ceil'

    If sum of p is less than 1 an extra group
    with excess elements are appended to the end.

    If sum of p is 1 any excess element from
    rounding is added to the last group. I.e.
    given 9 elements and p = [0.5,0.5] l_sizes
    will return 0,0,0,0,1,1,1,1,1.
    """
    # If p is passed as scalar
    # Convert to array
    if not hasattr(p, "__iter__"):
        # if not isinstance(p, list):
        p = [p]

    # Check if the sum of p is 0, 1 or in-between
    if False in [0 <= size <= 1 for size in p]:
        raise ValueError(
            'l_sizes: element of p is not equal to or within 0 and 1')

    # Check if p sums to more than 1
    if not 0 <= sum(p) <= 1:
        raise ValueError('l_sizes: sum of p is not equal to or within 0 and 1')

    # Is sum of p 1? Then we only want the number
    # of groups passed.
    if sum(p) == 1:
        fixed_n_groups = True
    else:
        fixed_n_groups = False

    if not rounding in ['floor', 'round', 'ceil']:
        raise ValueError(
            'l_sizes: rounding must be either floor, round, or ceil')

    # Find the sizes based on percentages given
    # Use rounding as specified by user
    if rounding == 'floor':
        sizes = np.floor([len(v) * size for size in p])
    elif rounding == 'round':
        sizes = np.round([len(v) * size for size in p])
    elif rounding == 'ceil':
        sizes = np.ceil([len(v) * size for size in p])

    # Find the remaining elements
    excess = len(v) - sum(sizes)

    # If any remaining elements
    if excess >= 0:
        if fixed_n_groups:
            sizes = np.append(sizes[:-1], sizes[-1:] + excess)
        else:
            # Add as size of last groups
            sizes = np.append(sizes, excess)

    # Check that we do not create more values than is in v
    if sum(sizes) != len(v):
        raise ValueError('l_sizes: Wrong number of elements in sizes')

    # Create grouping factor
    grouping_factor = [
        [group] * int(size)
        for size, group in zip(sizes, range(len(sizes)))
    ]

    # Flatten grouping factor
    grouping_factor = [item for sublist in grouping_factor for item in sublist]

    # If randomize is True, shuffle grouping factor
    if randomize:
        shuffle(grouping_factor)

    return grouping_factor



"""
@author: ludvigolsen
"""

import numpy as np
from random import shuffle


def _n_dist(v, n, randomize=False):
    """
    Creates grouping factor with
    distributed excess elements

    """
    len_v = float(len(v))

    divisor = len_v / n

    # Create range the length of v (+1 because it starts at 0)
    # Divide each element and get everything but the first element (0)
    v_divided = [vi / divisor for vi in range(int(len_v) + 1)][1:]

    # First round to a smaller decimal number to avoid ceil of
    # e.g., 7.00000000...03 becoming 8
    # Then round up to nearest integer
    v_ceiled = np.ceil(np.around(v_divided, 5)).astype(int)

    if randomize:
        shuffle(v_ceiled)

    return v_ceiled



import os
from copy import deepcopy
import pathlib
from itertools import combinations
from typing import Dict, List, Optional


# TODO Test (e.g., with "-" path included)
# TODO Add checks of path validity (bad symbols etc.) - Perhaps pathlib.Path does that?
# TODO allow checking if output is nested in input dir
# TODO Perhaps check compatibility on windows and other platforms
# TODO Convert docs to other format


def prepare_in_out_paths(
    named_collections: Dict[str, dict] = None,
    allow_none: bool = False,
    allow_overwriting: bool = True,
    allow_duplicates_in: List[str] = ["in_dirs", "out_dirs", "tmp_dirs"],
    disallowed_nestings: List[tuple] = [
        ("in_files", "tmp_dirs"),
        ("out_files", "tmp_dirs"),
        ("in_dirs", "tmp_dirs"),
        ("out_dirs", "tmp_dirs"),
    ],
    pathlib_out: bool = True,
    copy: bool = True,
):
    """
    Checks paths and converts them to `pathlib.Path` objects.
    Creates missing output directories.

    Note: Even when `allow_overwriting` is enabled,
    `in_files` and `out_files` cannot contain the same paths.

    Keys must be unique across the dictionaries.

    named_collections : dict of dicts
        A dict mapping collection name -> collection where a collection
        is a dict mapping (argument) name -> path.
        For the `in_files` collection, a path can also be "-" when streaming input.
            This value will remain a string and will not be checked
            for duplication, existence, etc.
    :param allow_none: Whether to allow paths (i.e. dict values) to be `None`,
        in which case they will be ignored.
    :param allow_overwriting: Whether to allow `out_files` paths
        to already exist.
    allow_duplicates_in : list
        List of collections to allow duplicate paths in. One of:
            {'in_files', 'in_dirs', 'out_files', 'out_dirs', 'tmp_files', 'tmp_dirs'}.
    disallowed_nestings : list of tuples
        Pairs of collections where the paths of the first cannot
        be nested inside the paths of the second.
        By default, the in and out paths cannot be located within
        the temporary directories, as these might be deleted.
    :param pathlib_out: Whether to convert the paths to
        `pathlib.Path` objects on return.
    :returns: Updated dicts (or `None`) for `in_files`, `in_dirs`, `out_files` and `out_dirs`.
        Paths are `pathlib.Path` objects that have been "resolved".
    """

    if copy:
        # Nested dictionaries -> deep copy
        named_collections = deepcopy(named_collections)

    # Check keys are unique

    # Assert non-overlapping keys in all combinations of collections
    for k1, k2 in set(combinations(named_collections.keys(), 2)):
        _check_different_keys(
            named_collections[k1], named_collections[k2], names=[k1, k2]
        )

    # Check paths in dicts

    # Check input filepaths
    named_collections["in_files"] = _check_paths_dict(
        d=named_collections["in_files"],
        d_name="in_files",
        allow_none=allow_none,
        check_duplicates="in_files" not in allow_duplicates_in,
        assert_exists=True,
        path_type="file",
    )

    # Check output filepaths
    named_collections["out_files"] = _check_paths_dict(
        d=named_collections["out_files"],
        d_name="out_files",
        allow_none=allow_none,
        check_duplicates="out_files" not in allow_duplicates_in,
        assert_exists=False,
        assert_missing=not allow_overwriting,
        path_type="file",
    )

    # Check tmp filepaths
    named_collections["tmp_files"] = _check_paths_dict(
        d=named_collections["tmp_files"],
        d_name="tmp_files",
        allow_none=allow_none,
        check_duplicates="tmp_files" not in allow_duplicates_in,
        assert_exists=False,
        assert_missing=not allow_overwriting,
        path_type="file",
    )

    # Check input directory paths
    named_collections["in_dirs"] = _check_paths_dict(
        d=named_collections["in_dirs"],
        d_name="in_dirs",
        allow_none=allow_none,
        check_duplicates="in_dirs" not in allow_duplicates_in,
        assert_exists=True,
        path_type="directory",
    )

    # Check output directory paths
    named_collections["out_dirs"] = _check_paths_dict(
        d=named_collections["out_dirs"],
        d_name="out_dirs",
        allow_none=allow_none,
        check_duplicates="out_dirs" not in allow_duplicates_in,
        path_type="directory",
    )

    # Check temporary directory paths
    named_collections["tmp_dirs"] = _check_paths_dict(
        d=named_collections["tmp_dirs"],
        d_name="tmp_dirs",
        allow_none=allow_none,
        check_duplicates="tmp_dirs" not in allow_duplicates_in,
        assert_missing=True,
        path_type="directory",
    )

    # Check for path duplicates across file collections
    # and make sure `tmp_dirs` are unique as they might be deleted after use
    to_check_for_dups = [
        ("in_files", "out_files"),
        ("in_files", "tmp_files"),
        ("out_files", "tmp_files"),
        ("tmp_dirs", "in_dirs"),
        ("tmp_dirs", "out_dirs"),
    ]
    for coll_1, coll_2 in to_check_for_dups:
        _check_duplicates_across_dicts(
            d1=named_collections[coll_1], d2=named_collections[coll_2]
        )

    # Check nestings
    # Paths in coll_1 cannot be nested within coll_2
    for coll_1, coll_2 in disallowed_nestings:
        _check_nestings(d1=named_collections[coll_1], d2=named_collections[coll_2])

    # Convert all paths to pathlib.Path objects
    if pathlib_out:
        for key, coll in named_collections.items():
            named_collections[key] = _normalize_paths(coll, type_fn=pathlib.Path)

    # Combine collections to single dict
    # for faster and simpler look-up
    # Hence the need for unique keys
    # across the dicts
    specified_collections = [
        coll for coll in named_collections.values() if coll is not None
    ]
    if not len(specified_collections):
        all_paths = {}
    for i, coll in enumerate(specified_collections):
        if i == 0:
            all_paths = coll.copy()
        else:
            all_paths.update(coll)

    return named_collections, all_paths


# Utilities


def _check_paths_dict(
    d: Optional[dict],
    d_name: str,
    allow_none: bool = False,
    check_duplicates: bool = True,
    assert_exists: bool = False,
    assert_missing: bool = False,
    path_type: str = "file",
):
    if assert_exists and assert_missing:
        raise ValueError("Both `assert_exists` and `assert_missing` were enabled.")

    if d is not None:
        if not (d is None or isinstance(d, dict)):
            raise ValueError(f"{d_name} must be either a `dict` of paths or `None`.")
        if allow_none:
            d = _rm_none_elements(d)
        _check_elements(d)
        # Convert to string
        d = _normalize_paths(d, type_fn=str)
        if check_duplicates:
            _check_duplicates(d)
        if assert_exists or assert_missing:
            _check_paths_exist(
                d,
                check_type=path_type,
                raise_when="unknown" if assert_exists else "known",
            )
    return d


def _rm_none_elements(d: dict):
    if d is not None:
        return {k: v for k, v in d.items() if v is not None}
    else:
        return None


def _normalize_paths(d: dict, type_fn=str):
    def path_formatter(v):
        if isinstance(v, str) and v == "-":
            return str(v)
        else:
            return type_fn(pathlib.Path(v).resolve())

    if d is not None:
        # Convert to full paths as strings
        return {k: path_formatter(v) for k, v in d.items()}
    return None


def _check_elements(d: dict):
    if d is not None:
        # Check values
        invalid_keys = [
            k for k, v in d.items() if not isinstance(v, (str, pathlib.PurePath))
        ]
        if len(invalid_keys) > 0:
            raise ValueError(
                f"The following paths were of an invalid type: {', '.join(invalid_keys)}."
            )

        # Check keys are strings
        invalid_keys = [k for k in d.keys() if not isinstance(k, str)]
        if len(invalid_keys) > 0:
            raise ValueError(
                f"The following keys were of an invalid type: {', '.join(invalid_keys)}. Keys must be strings."
            )


def _find_duplicate_value_to_keys(d: dict):
    unique_values = set([str(v) for v in d.values()])
    if "-" in unique_values:
        unique_values = unique_values.difference(set("-"))
    # This might be slow with large dictionaries
    value_to_keys = {
        uv: [k for k, v in d.items() if str(v) == uv] for uv in unique_values
    }
    duplicates = {v: ks for v, ks in value_to_keys.items() if len(ks) > 1}
    if len(duplicates) > 0:
        return duplicates
    return {}


def _check_duplicates(d: dict):
    if d is not None:
        if len(set(d.values())) != len(d.values()):
            # Note: `_find_duplicate_value_to_keys` returns path->keys mappings
            duplicates = _find_duplicate_value_to_keys(d).values()
            if len(duplicates) > 0:
                # Reason for extra check:
                #   Could have been multiple "-" paths
                #   which would have been okay
                raise ValueError(
                    "Found duplicate paths for some paths. "
                    "Duplicate paths had the following names: "
                    f"\n{duplicates}"
                )


def _check_paths_exist(d: dict, check_type: str = "file", raise_when: str = "unkown"):
    assert raise_when in ["unknown", "known"]
    assert check_type in ["file", "directory"]
    check_fn = os.path.isfile if check_type == "file" else os.path.isdir
    for k, v in d.items():
        if str(v) == "-":
            continue
        if raise_when == "unknown":
            assert check_fn(str(v)), f"`{k}` was not a {check_type}: {v}"
        else:
            assert not check_fn(str(v)), f"`{k}` is an existing {check_type}: {v}"


def _check_duplicates_across_dicts(d1: Optional[dict], d2: Optional[dict]):
    # TODO This requires documentation!

    if d1 is None or d2 is None:
        return None

    # Copy to ensure we don't alter
    # original dicts in parent scope
    d1 = d1.copy()
    d2 = d2.copy()

    # Deduplicate each dict internally

    # Find keys to remove
    d1_duplicates_keys_to_remove = [
        k for _, keys in _find_duplicate_value_to_keys(d1) for k in keys[1:]
    ]
    d2_duplicates_keys_to_remove = [
        k for _, keys in _find_duplicate_value_to_keys(d2) for k in keys[1:]
    ]

    # Remove the keys
    for k in d1_duplicates_keys_to_remove:
        del d1[k]
    for k in d2_duplicates_keys_to_remove:
        del d2[k]

    # Combine the two dictionaries
    combined = d1
    combined.update(d2)

    # Check for duplicate values across the two dictionaries
    _check_duplicates(combined)


def _check_different_keys(d1: Optional[dict], d2: Optional[dict], names: list):
    if d1 is not None and d2 is not None:
        key_intersection = list(set(d1.keys()).intersection(set(d2.keys())))
        # Raise error if any intersection
        if len(key_intersection):
            raise ValueError(
                f"The same keys were found in `{names[0]}` and `{names[1]}`: {key_intersection}."
            )


def _check_nestings(d1: Optional[dict], d2: Optional[dict]):
    if d1 is not None and d2 is not None:
        nestings = _find_nested(d1, d2)
        if nestings:
            nestings_string = [f"({x} in {y})" for x, y in nestings]
            raise ValueError(
                f"Found {len(nestings)} disallowed nested paths: "
                f"{', '.join(nestings_string)}."
            )


def _find_nested(d1, d2):
    """
    Find paths in d1 that are nested inside paths in d2.

    Returns tuples with keys `(d1_key, d2_key)` where `d1_path` was a subdirectory of `d2_path`.
    """
    # Copy to ensure we don't alter
    # original dicts in parent scope
    d1 = d1.copy()
    d2 = d2.copy()

    # Make paths `pathlib.Path` objects
    d1 = _normalize_paths(d1, type_fn=pathlib.Path)
    d2 = _normalize_paths(d2, type_fn=pathlib.Path)

    return [
        (d1_key, d2_key)
        for d1_key, d1_path in d1.items()
        for d2_key, d2_path in d2.items()
        if d2_path in d1_path.parents
    ]



"""
@author: ludvigolsen
"""

from typing import Optional, Union
from numbers import Number
import pandas as pd
from utipy.pandas import subset_by_levels
from .group_uniques import group_uniques
from .group import group


def fold(
    data: pd.DataFrame,
    n: Number = 5,
    id_col: Optional[str] = None,
    cat_col: Optional[str] = None,
    return_factor: bool = False,
    copy: bool = True
) -> Union[pd.DataFrame, pd.Series]:
    """
    Create balanced folds.
    Balance on a categorical column
    and/or make sure that datapoints that share 
    an ID (e.g., participant id) are kept in the
    same folds.

    """
    if copy:
        data = data.copy()

    # Create temporary index that we will
    # use to reorder the dataframe later
    data['.sorting_index'] = range(len(data))

    # If user specified id_col
    if id_col is not None:

        # And if user specified cat_col
        if cat_col is not None:

            # Subset data by levels of categorical column
            cat_subsets = subset_by_levels(data, cat_col)

            # Create groups for each subset based on unique values in id_col
            # Concatenate subsets with the new grouping factor
            data = pd.concat([group_uniques(data, n, id_col)
                             for data in cat_subsets])

        else:

            # Create groups based on unique values in id_col
            data = group_uniques(data, n, id_col, copy=False)

    else:

        # If user specified cat_col (but not id_col)
        if cat_col is not None:

            # Subset data by levels of categorical column
            cat_subsets = subset_by_levels(data, cat_col)

            # Group each subset and concatenate subsets
            data = pd.concat([group(data, n, cat_col) for data in cat_subsets])

        # If neither id_col or cat_col is specified
        else:

            # Group data by the first column in data
            # group() does this automatically if not
            # given a column.
            data = group(data, n, copy=False)

    # Sort data by the sorting index
    # and remove the sorting index
    data = data.sort_values('.sorting_index').drop('.sorting_index', 1)

    # If return factor is True
    if return_factor:

        # Only the return the grouping factor
        return data['group']

    else:

        # Return data
        return data



#!/usr/bin/env
# -*- coding: utf-8 -*-
"""
@author: ludvigolsen
"""

from typing import List
import pandas as pd
import numpy as np
import warnings

from utipy.utils.convert_to_df import convert_to_df


def polynomializer(
    data: pd.DataFrame,
    degree: int = 2,
    suffix: str = '_poly',
    exclude: List[str] = [],
    copy: bool = True
) -> pd.DataFrame:
    """
    Creates polymonial features.
    Adds suffix with information on which degree a column represents.


    Parameters
    ----------
    data : pd.DataFrame
        The data to add polynomial features to.
    degree : int
        How many degrees to add.
    suffix: str
        Text between column name and degree number.
        Added to the new columns.
    exclude: list
        List of column names to exclude, e.g., non-numeric
        columns.


    Returns
    -------
    pd.DataFrame with added polynomial features / columns.


    Examples
    --------

    >>> df = pd.DataFrame({'a': [1,2,3,4,5],
    ...                    'b': [2,3,4,5,6],
    ...                    'c': ['a','b','c','d','e']})
    >>> polynomializer(df, degree = 3, exclude = ['c'])

    """

    # Create copy of data
    if copy:
        data = data.copy()

    data, _ = convert_to_df(data)

    if exclude != []:
        cols_include = [c for c in data.columns if c not in exclude]
        data_include = data.filter(items=cols_include)
    else:
        data_include = data

    try:
        # Create dataframes with exponential columns
        polynomialized = [data_include ** deg for deg in range(degree + 1)[1:]]
    except:
        # This exception is most likely seen
        # because a column in data_include is NOT numeric.
        # So we exclude non-numeric columns, warn the user
        # and try again.
        numeric_data = data_include.select_dtypes(include=[np.number])

        # Get excluded columns and add to exclude
        auto_excluded = [
            i for i in data_include.columns
            if i not in numeric_data.columns
        ]
        exclude = np.concatenate([exclude, auto_excluded])

        if len(auto_excluded) != 0:
            warnings.warn(
                "Excluded {} non-numeric columns.".format(len(auto_excluded)))

            try:
                # Create dataframes with exponential columns
                polynomialized = [numeric_data **
                                  deg for deg in range(degree + 1)[1:]]
            except Exception as e:
                print("Something went wrong when creating polynomials.")
                raise e
        else:
            print("Something went wrong when creating polynomials.")
            raise

    # Function for adding suffix to column names
    def suffixicate(df, deg):
        if deg != 0:
            return df.add_suffix("{}{}".format(suffix, deg + 1))
        return df

    # Add suffices to dataframes
    polynomialized = map(
        lambda df, deg: suffixicate(df, deg),
        polynomialized,
        range(degree)
    )

    # Combine dataframes
    polynomialized = pd.concat(polynomialized, axis=1)

    # Combine processed data with excluded data
    data_all = pd.concat([data.filter(items=exclude), polynomialized], axis=1)

    # Reorder

    # First get all the new column names
    new_cols = [c for c in polynomialized.columns
                if c not in data.columns]

    # Append old and new column names
    all_columns_sorted = np.append(data.columns, new_cols)

    # Reorder dataframe
    data_ordered = data_all.filter(items=all_columns_sorted)

    return data_ordered



"""
@author: ludvigolsen
"""

from typing import List, Tuple, Union
import numpy as np
import pandas as pd
from utipy.utils.check_instance import check_instance
from utipy.utils.convert_to_type import convert_to_type

# TODO: Cythonize

def window(
        x: Union[list, np.ndarray, pd.Series], 
        size: int = 2, 
        gap: int = 1, 
        sample_rate: int = 1,
        rolling: bool = True, 
        reverse_direction: bool = False,
        discard_shorts: bool = True
    ) -> Tuple[List[np.ndarray], int]:
    """

    Splits array, e.g., time series, into rolling (optional) windows and returns as list of arrays and the number of windows.

    Parameters
    ----------
    x : list, np.ndarray, pd.Series
        The time series array to window.
    size : int
        Window size
    gap : int 
        Gap size.
    sample_rate : int
        Size and gap will be multiplied by the given sample rate
        allowing you to specify those in seconds instead of samples.
    rolling : bool
        Use rolling windows. 
        If False:
            Will grab "size * sample_rate" elements greedily.
            Be aware of the gap setting that defaults to 1.
    reverse_direction : bool
        Start from the end of the array instead of the beginning.
        Does not change order of elements within windows.
    discard_shorts: bool
        If the given array is shorter than size*sample_rate,
        return ([],0) if (True) or ([x],0) if (False).

    Returns
    -------
    List of np.ndarrays, number of windows

    """

    _ = check_instance(x)
    x = convert_to_type(x, 'np.ndarray')

    assert isinstance(size, int), "size must be an integer"
    assert isinstance(gap, int), "gap must be an integer"
    assert isinstance(sample_rate, int), "sample_rate must be an integer"
    assert isinstance(rolling, bool), "rolling must be a bool"
    assert isinstance(reverse_direction,
                      bool), "reverse_direction must be a bool"
    assert isinstance(discard_shorts, bool), "discard_shorts must be a bool"

    assert size >= 1, "size must be at least 1"
    assert sample_rate >= 1, "sample_rate must be at least 1"
    if rolling:
        assert gap >= 1, "gap must be at least 1 when creating rolling windows"

    # How many samples per file?
    n_samples = sample_rate * size
    gap_samples = gap * sample_rate

    # If the array is too short
    if len(x) < n_samples:
        if discard_shorts:
            return [], 0
        else:
            return [x], 0

    if rolling:
        n_windows = np.int32((len(x) - n_samples) / gap_samples) + 1
        if reverse_direction:
            stims = [x[len(x) - stimuli * gap_samples - n_samples:len(x) - stimuli * gap_samples]
                     for stimuli in range(n_windows)]
        else:
            stims = [x[stimuli * gap_samples:stimuli * gap_samples + n_samples]
                     for stimuli in range(n_windows)]
    else:
        modulus = len(x) % (n_samples + gap_samples)
        n_windows = (len(x) - modulus) / (n_samples + gap_samples)
        n_windows = np.int32(n_windows)

        # If we have enough samples for a window, just not the final gap
        if modulus >= n_samples:
            n_windows += 1

        if reverse_direction:
            stims = [x[len(x) - (stimuli + 1) * n_samples - stimuli * gap_samples:
                     len(x) - (stimuli) * n_samples - stimuli * gap_samples]
                     for stimuli in range(n_windows)]
        else:
            stims = [x[stimuli * n_samples + stimuli * gap_samples:
                     (stimuli + 1) * n_samples + stimuli * gap_samples]
                     for stimuli in range(n_windows)]

    return stims, n_windows





from typing import Callable, Optional, Tuple, Union
import numpy as np
import pandas as pd

from utipy.utils.messenger import Messenger, check_messenger


def print_nan_stats(
        x: Union[np.ndarray, pd.DataFrame],
        message: str,
        messenger: Optional[Callable] = Messenger(
            verbose=True, indent=0, msg_fn=print),
        indent: Optional[int] = None) -> None:
    """
    Print statistics about NaNs in an array.

    Parameters
    ----------
    x : `numpy.ndarray` or `pandas.DataFrame`
        The array / data frame to count NaNs in.
    message : str
        The message prior to the stats. Full message becomes:
            `indentation + message + ": " + num NaNs (percentage)` 
    messenger : `utipy.Messenger` or None
        A `utipy.Messenger` instance used to print/log/... information.
        When `None`, no printing/logging is performed.
        The messenger determines the messaging function (e.g., `print` or `log.info`)
        and indentation when `indent` is `None`.
    indent : int
        Indentation of message. When `None`, indentation is determined by `messenger`.
    """
    messenger = check_messenger(messenger)
    num_nans, perc = nan_stats(x)
    messenger(f"{message}: {num_nans} ({perc}%)", indent=indent)


def nan_stats(x: Union[np.ndarray, pd.DataFrame]) -> Tuple[int, float]:
    """
    Get statistics about NaNs in an array.

    Parameters
    ----------
    x : `numpy.ndarray`
        A numpy array.

    Returns
    -------
    int
        Number of NaNs in `x`.
    float
        Percentage of `x` that was NaN. 
        Between 0 and 100.
    """
    num_nans = np.count_nonzero(np.isnan(x))
    perc = np.round((num_nans / float(x.size)) * 100, decimals=2)
    return num_nans, perc




import pandas as pd


def move_column_inplace(df: pd.DataFrame, col: str, pos: int) -> None:
    """
    Move a column to a given column-index position.

    Parameters
    ----------
    df : `pandas.DataFrame`.
    col : str
        Name of column to move.
    pos : int
        Column index to move `col` to.
    """
    assert 0 <= pos < len(df.columns), \
        f"`pos` must be between 0 (incl.) and the number of columns -1. Was {pos}."
    col = df.pop(col)
    df.insert(pos, col.name, col)



"""
@author: ludvigolsen
"""

import pandas as pd
import numpy as np
from random import sample
from typing import Optional, List

from .resemble import resemble
from utipy.array.blend import blend
from utipy.utils.convert_to_df import convert_to_df

# TODO The label column concept is not properly described


def distort(
    data: pd.DataFrame,
    distribution: str = 'uniform',
    amount: float = 1.,
    size: float = 1.,
    randomize_original: bool = False,
    exclude: Optional[List[str]] = None,
    label_column: Optional[str] = None,
    keep_labels: bool = True,
    new_label: str = 'noise',
    append: bool = False
) -> pd.DataFrame:
    """
    Distort data in pandas DataFrame

    Generates data that resembles the original data by using descriptors
    of each column to create and sample from a specified distribution.
    This new data is blended with the original data to taste.


    Parameters
    ----------
    data : pd.DataFrame
        The data to distort.
    distribution : str
        Distribution to sample from.
            'uniform'
                between min. and max.
            'gaussian'
                from mean and std.
            'robust gaussian'
                from median and IQR.
            'poisson'
                NOT IMPLEMENTED PROPERLY YET.
            'shuffle'
                shuffles original data.
    amount : float
        Blend rate. Amount of generated data to keep.
        Percentage between 0-1
            0: Keep only original data.
            1: Keep only generated data.
            0.1: 10% generated / 90% original.
        A value in-between 0-1 will result in integers becoming floats.
    size : float
        Size of data to return relative to original dataframe.
        Percentage between 0-1.
        Observations are randomly sampled.
    randomize_original : bool
        Should the values of the inputted data be randomized?
    exclude : list of strings
        Names of columns not to generate.
        Are filled with NaN.
    label_column : str
        Name of column with labels.
        Used when having a categorical
        column that should either keep
        its labels or get a new label.
    keep_labels : bool
        Leave label column untouched.
    new_label : str
        Label to fill label column with.
        Used when keep_labels is False.
    append : bool
        Append output to original data


    Returns
    -------
    pd.DataFrame
        Distorted data. Either by itself or appended to original data.
    """

    # Check inputs
    # `exclude`` must be array not scalar

    # If data is a pd.Series or np.ndarray
    # Make into a dataframe and
    # remember what type it originally was
    data, _ = convert_to_df(data)

    # Subsets
    # Select excluded and included columns
    if exclude is not None:
        # Select column names to regenerate as noise
        included_cols = [
            col for col in data.columns
            if col not in np.append(label_column, exclude)
        ]
    else:
        # Select columns to regenerate as noise
        included_cols = data.columns[data.columns != label_column]

    # Get included columns from column names
    data_included = data.filter(items=included_cols)

    # Regenerate included columns as noise
    data_regenerated = data_included.apply(
        resemble,
        distribution=distribution
    )

    # Blend
    # Based on amount, blend the two signals
    if amount != 1:
        data_blended = pd.concat([
            blend(data[v], data_regenerated[v], amount=amount)
            for v in data_regenerated.columns
        ], axis=1)
    else:
        data_blended = data_regenerated

    # Label column
    # TODO Improve the above comment xD
    if label_column is not None:
        if keep_labels:
            data_blended[label_column] = data[label_column]
        else:
            # Add label column with the new label
            data_blended[label_column] = new_label

    # Set excluded columns to NaN
    if exclude is not None:
        # Create empty dataframe the shape of [rows, excluded cols]
        # , columns = exclude)
        null_data = pd.DataFrame(np.zeros([len(data), len(exclude)]))

        # Set 0'es to NaN
        null_data[null_data == 0] = 'nan'

        # Set column names
        null_data.columns = exclude

        # Add columns to the blended dataset
        data_blended = pd.concat([data_blended, null_data], axis=1)

    # Reorder
    # Get columns in original order
    data_ordered = data_blended.filter(items=data.columns)

    # Cut to 'size'
    if randomize_original:
        # Sample indices from range 0: n rows
        keep_indices = sample(range(0, len(data)), int(len(data) * size))

        # Get the rows with the sampled indices
        keep_data = data_ordered.iloc[keep_indices]
    else:
        keep_data = data_ordered[:int(len(data) * size)]

    # Append
    if append:
        # Make sure they are the same size
        # TODO Shouldn't it raise an error instead if they're not?
        return pd.concat([data[:len(keep_data)], keep_data])
    else:
        return keep_data



"""
@author: ludvigolsen
"""

from numbers import Number
import pandas as pd
from .methods.l_sizes import _l_sizes
from .methods.n_dist import _n_dist


def group_uniques(
    data: pd.DataFrame,
    n: Number,
    col: str,
    method: str = 'n_dist',
    copy: bool = True
) -> pd.DataFrame:
    """
    Add grouping factor to given
    dataframe by unique values in col.

    """
    if copy:
        data = data.copy()

    # Get unique IDs
    uniques = list(set(data[col]))

    # Create grouping factor from unique IDs
    if method == 'n_dist':
        all_group_ids = _n_dist(uniques, n, randomize=True)
    elif method == 'l_sizes':
        all_group_ids = _l_sizes(uniques, n, randomize=True)

    # Create dictionary with unique values
    # and groups
    groups_dict = {}
    groups_dict[col] = uniques
    groups_dict['group'] = all_group_ids

    # Create dataframe with unique IDs and the
    # grouping factor
    df_grouped = pd.DataFrame(groups_dict)

    # Merge and return the given dataframe with the
    # new grouping factor
    return data.merge(df_grouped, on=col, how='outer')



"""
@author: ludvigolsen
"""

from typing import Any, Union
import numpy as np
import pandas as pd

def convert_to_type(data: Any, data_type: str) -> Union[list, tuple, np.ndarray, pd.Series, pd.DataFrame]:
    
    """
    Converts (if necessary) data into either of these data types:
     - list
     - tuple
     - np.ndarray
     - pd.Series
     - pd.DataFrame
    
    """
    if data_type not in ["list", "tuple", "np.ndarray", "pd.Series", "pd.DataFrame"]:
        raise ValueError(f"`data_type` was unknown: '{data_type}'.")

    # Convert 
    if data_type == 'list' and not isinstance(data, list):
        return list(data)
    elif data_type == 'tuple' and not isinstance(data, tuple):
        return tuple(data)
    elif data_type == 'np.ndarray' and type(data).__module__ != np.__name__:
        return np.asarray(data)    
    elif data_type == 'pd.Series' and not isinstance(data, pd.Series):
        return pd.Series(data)
    elif data_type == 'pd.DataFrame' and not isinstance(data, pd.DataFrame):
        try:
            return pd.DataFrame(data)
        except:
            return pd.DataFrame({'x':data})
    return data
        



import pathlib
import shutil
from typing import Callable, Optional, Union

from utipy.utils.messenger import Messenger, check_messenger


def mk_dir(
    path: Union[str, pathlib.Path],
    arg_name: Union[str, None] = "",
    raise_on_exists: bool = False,
    messenger: Optional[Callable] = Messenger(
        verbose=True, indent=0, msg_fn=print)
):
    """
    Make directory if it doesn't exist.

    Parameters
    ----------
    path : str or `pathlib.Path`
        Path to directory to make.
    arg_name : str or None
        Name of path argument/variable for message 
        when creating a directory and `messenger.verbose` is `True`.
    raise_on_exists : bool
        Whether to raise a `FileExistsError` when the directory already exists.
    messenger : `utipy.Messenger` or None
        A `utipy.Messenger` instance used to print/log/... information.
        When `None`, no printing/logging is performed.
        The messenger determines the messaging function (e.g., `print`)
        and potential indentation.
    """
    path = pathlib.Path(path)
    path_exists = path.exists()

    # Prepare arg name
    arg_name = _prep_arg_name(arg_name)

    # Check messenger (always returns Messenger instance)
    messenger = check_messenger(messenger)

    # Fail for existing directory (when specified)
    # Or exit function
    if path_exists:
        if raise_on_exists:
            raise FileExistsError(
                f"{arg_name}directory already exists: {path.resolve()}")
        return

    # Message user about the creation of a new directory
    messenger(
        f"{arg_name}directory does not exist and will be created: "
        f"{path.resolve()}"
    )

    # Create new directory if it does not already exist
    try:
        path.mkdir(parents=True, exist_ok=not raise_on_exists)
    except FileExistsError as e:
        # In this case, the directory was likely created between
        # our existence check and our creation attempt
        if raise_on_exists:
            raise FileExistsError(
                f"{arg_name}directory already exists: {path.resolve()}")


def rm_dir(
    path: Union[str, pathlib.Path],
    arg_name: Union[str, None] = "",
    raise_missing: bool = False,
    raise_not_dir: bool = True,
    shutil_ignore_errors: bool = False,
    shutil_onerror: Optional[Callable] = None,
    messenger: Optional[Callable] = Messenger(
        verbose=True, indent=0, msg_fn=print)
):
    """
    Remove directory and its contents if it exists using `shutil.rmtree()`.

    Parameters
    ----------
    path : str or `pathlib.Path`
        Path to directory to remove.
    arg_name : str or None
        Name of path argument/variable for message 
        when creating a directory and `messenger.verbose` is `True`.
    raise_missing : bool
        Whether to raise a RuntimeError when the directory does not exist.
    raise_not_dir : bool
        Whether to raise a RuntimeError when the path is not to a directory.
    shutil_ignore_errors : bool
        Passed to the `ignore_errors` argument in `shutil.rmtree()`.
    shutil_onerror : bool
        Passed to the `onerror` argument in `shutil.rmtree()`.
    messenger : `utipy.Messenger` or None
        A `utipy.Messenger` instance used to print/log/... information.
        When `None`, no printing/logging is performed.
        The messenger determines the messaging function (e.g., `print`)
        and potential indentation.
    """
    path = pathlib.Path(path)
    path_exists = path.exists()

    # Prepare arg name
    arg_name = _prep_arg_name(arg_name)

    # Check messenger (always returns Messenger instance)
    messenger = check_messenger(messenger)

    if raise_missing and not path_exists:
        raise RuntimeError(f"{arg_name}path did not exist: {path}")

    if path_exists and raise_not_dir and not path.is_dir():
        raise RuntimeError(f"{arg_name}path was not a directory: {path}")

    if path_exists and path.is_dir():
        # Message user about the removal of the directory
        messenger(
            f"{arg_name}directory will be removed: "
            f"{path.resolve()}"
        )
        shutil.rmtree(path, ignore_errors=shutil_ignore_errors,
                      onerror=shutil_onerror)


def _prep_arg_name(arg_name):

    if arg_name is None or not arg_name:
        arg_name = ""
    else:
        arg_name = f"`{arg_name}` "
    return arg_name



# pandas __init__.py

from .drop import drop
from .makes_up import makes_up
from .resemble import resemble
from .distort import distort
from .subset_by_levels import subset_by_levels
from .polynomializer import polynomializer



from contextlib import contextmanager
from typing import Callable, Union

from utipy.string.random_strings import random_alphanumeric

from .timestamps import Timestamps


class StepTimer(Timestamps):
    def __init__(
        self, message: str = "Took:", verbose: bool = True, msg_fn: Callable = print
    ) -> None:
        """
        A `StepTimer` can be used in `with` statements
        to time steps of code and print the execution time.
        All timestamps are kept, why we can get the
        total time in the end as well.

        See `Timestamps` for methods that can be applied to recorded timestamps.

        Parameters
        ----------
        message : str
            Default message to prefix time with.
            Can be overridden for separate timings with the `message`
            argument in `.time_step()`.
        verbose : bool
            Whether to print/log/.. the time a step took when using `.time_step()`.
        msg_fn : callable
            The function to use for printing/logging the message.
            E.g., `print` or `logging.info`.
        """
        super().__init__()
        self.message = message
        self.verbose = verbose
        self.msg_fn = msg_fn

    @contextmanager
    def time_step(
        self, indent: int = 4, message: Union[str, None] = None, name_prefix=None
    ):
        """
        Function to use in `with` statement. E.g.:

          > timer = StepTimer()
          > with timer.time_step():
          >     a = 2
          >     b = a * 4
          output: 'Took: 00:00:01'

        Can be used in a nested fashion.

        Parameters
        ----------
        indent : int
            How many spaces to indent message.
        message : str or None
            Step-specific prefix to the time string.
            When `None`, the message supplied during
            inialization is used.
        name_prefix : str or None
            Prefix to record timestamp with.
                The initial timestamp will be recorded with the name `name_prefix + '_start'`.
                The final timestamp will be recorded with the name `name_prefix + '_end'`.
            This allows easily getting the specific timepoints with `.get_stamp()` or
            the difference between two stamps with `.took()`.
            When not specified, a prefix is generated.

        Yields
        ------
        `None` once.

        Prints
        ------
        When `self.verbose` is `True`,
            prints message + formatted time.
        """
        assert indent >= 0
        if not name_prefix:
            name_prefix = f"step_{len(self)}_{random_alphanumeric(size=5)}"
        try:
            self.stamp(name=name_prefix + "_start")
            yield None
        finally:
            self.stamp(name=name_prefix + "_end")
            mess = self.message if message is None else message
            if self.verbose:
                self._print_runtime(
                    indent=indent,
                    message=mess,
                    start=name_prefix + "_start",
                    end=name_prefix + "_end",
                )

    def _print_runtime(
        self,
        indent: int,
        message: str,
        start: Union[int, str] = -2,
        end: Union[int, str] = -1,
    ) -> None:
        """
        Prints the indented message and step time.
        """
        indent_str = "".join([" " for _ in range(indent)])
        self.msg_fn(f"{indent_str}{message} {self.took(start=start, end=end)}")




# TODO Add tests

class StepIdx:

    def __init__(self, start_idx: int = 0) -> None:
        """
        A counter for stepping through some list of tasks.
        Get current, previous and next indices.

        Parameters
        ----------
        start_idx : int
            Starting index.
        """
        self.current = start_idx

    def step(self) -> None:
        """
        Take a step. 
        Increases "current index" by 1.
        """
        self.current += 1

    @property
    def previous(self) -> int:
        """
        Get the index of the previous step.

        Returns
        -------
        int
            Previous index.
        """
        if self.current == 0:
            raise ValueError("`current` was 0 which has no `previous` index.")
        return self.current - 1

    @property
    def next(self) -> int:
        """
        Get the index of the following step.

        Returns
        -------
        int
            Next index.
        """
        return self.current + 1



"""
@author: ludvigolsen
"""

from typing import Union
import numpy as np
import pandas as pd
from utipy.utils.check_instance import check_instance
from utipy.utils.convert_to_type import convert_to_type


def blend(
        x1: Union[list, np.ndarray, pd.Series], 
        x2: Union[list, np.ndarray, pd.Series], 
        amount: float = 0.5
    ) -> Union[list, np.ndarray, pd.Series]:
    """
    Blend two arrays

    Parameters
    ----------
    x1 : list, np.ndarray, pd.Series
        The first array.
    x2 : list, np.ndarray, pd.Series
        The second array.
    amount : float
        Blend rate. 
        Percentage between 0-1
            0: Keep only x1.  
            1: Keep only x2.  
            0.1: 10% x2 / 90% x1.  
        A value in-between 0-1 will result in integers becoming floats.

    Returns
    -------
    list, np.ndarray, pd.Series
        Blended array with type of the original (x1)

    Examples
    --------
    
    >>> x1 = [1,2,3,4,5]  
    >>> x2 = [4,5,6,7,8]  
    >>> blend(x1, x2, amount = 0.5)  
    [2.5,3.5,4.5,5.5,6.5]

    """

    # Get instance types (np.ndarray, list, pd.Series)
    instance_type = check_instance(x1)

    x1_weighted = np.multiply(x1, (1 - amount))
    x2_weighted = np.multiply(x2, amount)

    blended = x1_weighted + x2_weighted

    # Convert to original type (np.ndarray, list, pd.Series)
    return convert_to_type(blended, instance_type)



from string import ascii_lowercase, ascii_uppercase
import itertools
from typing import List, Optional


def letter_strings(
    n: int,
    num_chars: Optional[int] = None,
    upper: bool = False,
    descending: bool = False,
) -> List[str]:
    """
    Generate a sequence of `n` letter strings (aa, ab, ac, ...).

    Parameters
    ----------
    n : int
        Number of letter strings to generate.
        When `num_chars` is not specified, this also determines the number of
        characters in the strings.
    num_chars : int or None
        The (starting) number of characters in the strings.
        When `None`, this is determined by `n`.
    upper : bool
        Whether to use uppercase letters instead of lowercase.
    descending : bool
        Whether to use letters in descending order.

    Returns
    -------
    list of str
        A list of letter strings.
        E.g., `['aa', 'ab', 'ac']`.
    """

    # Number of letters in the ascii alphabet
    num_available_letters = 26

    if num_chars is None:
        # Find the necessary number of characters
        # to create the IDs
        num_chars = 1
        while n > num_available_letters**num_chars:
            num_chars += 1

    # Init letter string generator
    generator = letter_string_generator(
        num_chars=num_chars, upper=upper, descending=descending
    )

    # Get the n first letter IDs
    return list(itertools.islice(generator, n))


def letter_string_generator(
    num_chars: int = 1, upper: bool = False, descending: bool = False
):
    """
    Letter string generator.

    Parameters
    ----------
    num_chars : int
        The (starting) number of characters in the yielded strings.
    upper : bool
        Whether to use uppercase letters instead of lowercase.
    descending : bool
        Whether to use letters in descending order.

    Yields
    ------
    str
        Given `num_chars` is 2:
            "aa", "ab", "ac", ... "aaa", "aab" ...
    """
    letters = ascii_uppercase if upper else ascii_lowercase
    if descending:
        letters = letters[::-1]
    while True:
        for s in itertools.product(letters, repeat=num_chars):
            yield "".join(s)
        num_chars += 1



#!/usr/bin/env
# -*- coding: utf-8 -*-
"""
@author: ludvigolsen
"""

from typing import List
import pandas as pd

# TODO Could this not be easily done with group_by -> split?
# Consider refactoring or removing this function


def subset_by_levels(
    data: pd.DataFrame,
    cat_col: str,
    drop_cat_col: bool = False
) -> List[pd.DataFrame]:
    """
    Subsets dataframe by each level of a categorical column.


    Parameters
    ----------
    data : pd.DataFrame
        The dataframe to subset.
    cat_col : str, int
        Name or index of categorical column to subset by.
    drop_cat_col: bool
        Remove the categorical column from each subset.


    Returns
    -------
    list of pd.DataFrames


    Examples
    --------

    >>> df = pd.DataFrame({'a': [1,2,3,4],
    ...                    'b': [2,3,4,5],
    ...                    'c': ['a','b','a','b']})
    >>> subset_by_levels(df, cat_col = 'c')

    """

    # Find the unique values (levels)
    # in the given categorical column
    levels = list(set(data[cat_col]))

    # Create subsets for each level
    # Saved as a list of dataframes
    subsets = [pd.DataFrame(data[data[cat_col] == l]) for l in levels]

    # Remove cat_col
    if drop_cat_col:
        subsets = [d.drop(cat_col, axis=1) for d in subsets]

    # Return subsets
    return subsets



"""
@author: ludvigolsen
"""

from typing import Union
import pandas as pd
import numpy as np
from utipy.measures.iqr import _iqr

# For pandas objects

def _extended_describe(x:Union[pd.Series, np.ndarray, list]) -> dict:
    """
    Adds median and IQR to `pandas.Series.describe()`.
    Only works for numeric series.
    """
    desc = pd.Series(x).describe()
    desc['median'] = np.median(x)
    desc['IQR'] = _iqr(x)
    return dict(desc)
    


"""
@author: ludvigolsen
"""

from typing import Any, Tuple
import numpy as np
import pandas as pd


def convert_to_df(data: Any) -> Tuple[pd.DataFrame, str]:
    """
    Checks the type of `data`.
    If it is not a `pd.DataFrame` it 
    attempts to convert to pd.DataFrame

    """
    data_type = 'pd.DataFrame'
    if isinstance(data, pd.DataFrame):
        return data, data_type

    elif isinstance(data, pd.Series):
        data_type = 'pd.Series'
        data = pd.DataFrame({'x': data})

    elif type(data).__module__ == np.__name__:
        try:
            data = pd.DataFrame(data)
            data_type = 'np.ndarray'

        except:
            raise TypeError(f"`data` is the wrong format: '{type(data)}'.")
    else:
        raise TypeError("`data` is the wrong format.")

    return data, data_type



"""
@author: ludvigolsen
"""

from typing import Callable, Optional, Union, List
import pandas as pd
from numbers import Number

from utipy.utils.messenger import Messenger, check_messenger
from .makes_up import makes_up


def drop(
    data: pd.DataFrame,
    value: Union[str, Number] = 'NaN',
    thresh: float = 0,
    direction: str = '>',
    axis: int = 0,
    include: Optional[List[str]] = None,
    exclude: Optional[List[str]] = None,
    copy: bool = True,
    messenger: Optional[Callable] = Messenger(
        verbose=True, indent=0, msg_fn=print)
) -> pd.DataFrame:
    """
    Drop rows or columns from pandas DataFrame based on values.

    Drop rows / columns if specific value, or any value, is represented too much, too little, etc.

    Commands:
        'Drop [axis] if [value] appears [direction] than [thresh] percent of the time.'
    E.g.:
        'Drop columns if 0 appears more than 90 percent of the time.'
        'Drop rows if *any* value appears exactly 77 percent of the time.'


    Parameters
    ----------
    data : pd.DataFrame
        The data to distort.
    value : str / int / float
        The value to match.
            Regular value, 
            'any',
            'NaN',
            'inf'
    thresh : float
        Threshold.
        Percentage between 0-1.
    direction : str
        Operator sign for comparison.
            '>', '<', '>=', '<=', '=='.
    axis : int
        0 for columns, 1 for rows.
    include : list of strings
        Names of columns / indices of rows to search within. 
        None means ALL are included unless otherwise specified, see *exclude*.
    exclude : list of strings
        Names of columns / indices of rows NOT to search within.
        None means no columns/rows are excluded unless otherwise specified, see *include*.
    messenger : `utipy.Messenger` or None
        A `utipy.Messenger` instance used to print/log/... information.
        When `None`, no printing/logging is performed.
        The messenger determines the messaging function (e.g., `print`)
        and potential indentation.


    Returns
    -------
    pd.DataFrame


    Examples
    --------

    Remove all rows with any NaNs in dependent variable 'y'
    >>> drop(data, value = 'NaN', axis = 1, thresh = 0, 
    ...      direction = '>', cols = ['y'])

    Remove all columns with only 1 unique value.
    I.e. the same value in 100% of the rows.
    >>> drop(data, value = 'any', axis = 0, thresh = 1, 
    ...      direction = '==')

    Remove all columns that have less than 30% NaNs
    >>> drop(data, value = 'NaN', axis = 0, thresh = 0.3, 
    ...      direction = '<')

    """

    # Check messenger (always returns Messenger instance)
    messenger = check_messenger(messenger)

    if value is None:
        raise ValueError('value cannot be None.')

    if axis not in [0, 1]:
        raise ValueError("`axis` must be 0 or 1")

    # Create copy of dataframe
    if copy:
        data = data.copy()

    if exclude is not None and include is not None:
        raise ValueError("Either include or exclude must be None.")

    # Columns
    if axis == 0:

        if exclude is not None:
            # Create include list
            include = [col for col in data.columns if col not in exclude]

        if include is not None:
            # Subset dataframe to only work on included cols
            data_cols = data.filter(items=include)

            # Find columns / rows to drop
            to_drop = _find_exceeders(data_cols, value,
                                      thresh, direction,
                                      axis=axis)
        else:
            # Find columns / rows to drop
            to_drop = _find_exceeders(data, value, thresh,
                                      direction, axis=axis)

        # Drop columns
        messenger(f'Dropped {len(to_drop)} columns.')

        return data.drop(to_drop, axis=1)

    # Rows
    elif axis == 1:
        # Find columns / rows to drop
        to_drop = _find_exceeders(data, value, thresh,
                                  direction, axis=axis)

        # Remove indices not in include or in exclude
        # TODO use sets instead
        if exclude is not None:
            to_drop = [i for i in to_drop if i not in exclude]
        elif include is not None:
            to_drop = [i for i in to_drop if i in include]

        # Drop rows
        messenger(f'Dropped {len(to_drop)} rows.')

        return data.drop(data.index[to_drop], axis=0)


# Wrapper for calling makes_up - for finding columns / rows to drop
def _find_exceeders(data, value, thresh, direction, axis):
    """Internal wrapper function for calling makes_up()"""

    # Notice:
    # pd apply() passes the Series as objects, even
    # when it's the columns it's working with!

    if axis not in [0, 1]:
        raise ValueError("`axis` must be 0 or 1")

    exceeders = data.apply(
        makes_up,
        axis=axis,
        value=value,
        thresh=thresh,
        direction=direction
    )

    if axis == 0:
        # Find columns to drop
        to_drop = [
            col for col, exceedes in
            zip(data.columns, exceeders)
            if exceedes
        ]

    elif axis == 1:
        # Find rows to drop
        to_drop = [
            row for row, exceedes in
            zip(data.index, exceeders)
            if exceedes
        ]

    return to_drop
